import numpy as np
import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.classification.f1_score import F1Score


@pytest.fixture()
def fixture_metric():
    return load_metric("f1_score")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "f1_score" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, F1Score)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_integer_labels(fixture_metric):
    """That can use integer labels."""
    metric = fixture_metric
    predictions = [0, 1, 0, 0, 0, 1]
    references = [0, 1, 1, 0, 0, 1]
    result = metric.compute(predictions=predictions, references=references)
    assert result["f1_score"] == 0.8


def test_multiclass_macro_average(fixture_metric):
    """Multiclass case with unweighted mean average across classes."""
    metric = fixture_metric
    predictions = [0, 2, 1, 0, 0, 1]
    references = [0, 1, 2, 0, 1, 2]
    result = metric.compute(
        predictions=predictions,
        references=references,
        average="macro",
    )
    assert pytest.approx(result["f1_score"], 0.01) == 0.267


def test_multiclass_micro_average(fixture_metric):
    """Multiclass case with metrics calculated globally."""
    metric = fixture_metric
    predictions = [0, 2, 1, 0, 0, 1]
    references = [0, 1, 2, 0, 1, 2]
    result = metric.compute(
        predictions=predictions,
        references=references,
        average="micro",
    )
    assert pytest.approx(result["f1_score"], 0.01) == 0.333


def test_multiclass_weighted_average(fixture_metric):
    """Multiclass case with average across classes weighted by support."""
    metric = fixture_metric
    predictions = [0, 2, 1, 0, 0, 1]
    references = [0, 1, 2, 0, 1, 2]
    result = metric.compute(
        predictions=predictions,
        references=references,
        average="weighted",
    )
    assert pytest.approx(result["f1_score"], 0.01) == 0.267


def test_multiclass_none_average(fixture_metric):
    """Multiclass case with scores returned for each class."""
    metric = fixture_metric
    predictions = [0, 2, 1, 0, 0, 1]
    references = [0, 1, 2, 0, 1, 2]
    result = metric.compute(
        predictions=predictions,
        references=references,
        average=None,
    )
    np.testing.assert_allclose([0.8, 0, 0], result["f1_score"])


def test_perfect_score(fixture_metric):
    """That get unitary score for perfect predictions."""
    metric = fixture_metric
    predictions = [0, 1, 1, 0, 0, 1]
    references = [0, 1, 1, 0, 0, 1]
    result = metric.compute(predictions=predictions, references=references)
    assert result["f1_score"] == 1.0


@pytest.mark.xfail
def test_binary_multilabel():
    """Multilabel case with binary label indicators."""
    metric = load_metric("f1_score", config_name="multilabel")
    predictions = [[0, 0, 0], [1, 1, 1], [1, 1, 0]]
    references = [[0, 0, 0], [1, 1, 1], [0, 1, 1]]
    result = metric.compute(predictions=predictions, references=references)
    np.testing.assert_allclose([2 / 3, 1, 2 / 3], result["f1_score"])



import numpy as np
import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.classification.average_precision import AveragePrecision

metric_name = "average_precision"


@pytest.fixture()
def fixture_metric():
    return load_metric(metric_name)


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert metric_name in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, AveragePrecision)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_binary_classification(fixture_metric):
    """That can score binary classification."""
    metric = fixture_metric
    references = [1, 0, 1, 1, 0, 0]
    predictions = [0.5, 0.2, 0.99, 0.3, 0.1, 0.7]
    result = metric.compute(references=references, predictions=predictions)
    assert metric_name in result
    assert pytest.approx(result[metric_name], 0.01) == 0.81


def test_multilabel_classification():
    """That can score multilabel classification."""
    metric = load_metric(metric_name, config_name="multilabel")
    references = [[1, 1, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [0, 1, 1], [1, 0, 1]]
    predictions = [
        [0.3, 0.5, 0.2],
        [0.7, 0.2, 0.1],
        [0.005, 0.99, 0.005],
        [0.2, 0.3, 0.5],
        [0.1, 0.1, 0.8],
        [0.1, 0.7, 0.2],
    ]
    result = metric.compute(
        references=references,
        predictions=predictions,
        average=None,
    )
    assert metric_name in result
    assert np.allclose(result[metric_name], [0.87, 0.73, 0.92], rtol=0, atol=1e-2)



from molflux.metrics.load import load_suite
from molflux.metrics.metric import Metrics
from molflux.metrics.suites.catalogue import list_suites

suite_name = "classification"


def test_in_catalogue():
    """That the suite is in the catalogue."""
    suites = list_suites()
    assert suite_name in suites


def test_loads():
    """That the suite can be loaded without errors."""
    assert load_suite(name=suite_name)


def test_is_metrics():
    """That the suite returns a metrics collection."""
    suite = load_suite(name=suite_name)
    assert isinstance(suite, Metrics)



from dataclasses import asdict
from typing import Any, Literal

from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.models.pyod import (
    PyODClassificationMixin,
    PyODModelBase,
    PyODModelConfig,
)

try:
    from pyod.models.hbos import HBOS
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("pyod", e) from None


_DESCRIPTION = """
Histogram- based outlier detection (HBOS) is an efficient unsupervised
method. It assumes the feature independence and calculates the degree
of outlyingness by building histograms. See :cite:`goldstein2012histogram`
for details.

Two versions of HBOS are supported:
- Static number of bins: uses a static number of bins for all features.
- Automatic number of bins: every feature uses a number of bins deemed to
  be optimal according to the Birge-Rozenblac method
  (:cite:`birge2006many`).
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
n_bins : int or string, optional (default=10)
    The number of bins. "auto" uses the birge-rozenblac method for
    automatic selection of the optimal number of bins for each feature.

alpha : float in (0, 1), optional (default=0.1)
    The regularizer for preventing overflow.

tol : float in (0, 1), optional (default=0.5)
    The parameter to decide the flexibility while dealing
    the samples falling outside the bins.

contamination : float in (0., 0.5), optional (default=0.1)
    The amount of contamination of the data set,
    i.e. the proportion of outliers in the data set. Used when fitting to
    define the threshold on the decision function.
"""

BinningMethod = Literal["auto"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class HBOSDetectorConfig(PyODModelConfig):
    n_bins: int | BinningMethod = 10
    alpha: float = 0.1
    tol: float = 0.5
    contamination: float = 0.1


class HBOSDetector(PyODClassificationMixin, PyODModelBase[HBOSDetectorConfig]):
    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    @property
    def _config_builder(self) -> type[HBOSDetectorConfig]:
        return HBOSDetectorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> HBOS:
        config = self.model_config
        return HBOS(
            n_bins=config.n_bins,
            alpha=config.alpha,
            tol=config.tol,
            contamination=config.contamination,
        )



from dataclasses import dataclass
from typing import Any

from molflux.modelzoo.protocols import Model


@dataclass
class ModelArtefact:
    name: str
    tag: str
    config: dict[str, Any]
    model: Model


def pack(model: Model) -> ModelArtefact:
    name = model.name
    tag = model.tag
    config = model.config
    model = model

    return ModelArtefact(
        name=name,
        tag=tag,
        config=config,
        model=model,
    )


def unpack(artefact: ModelArtefact) -> Model:
    model = artefact.model
    return model



from molflux.splits.utils import partition


def test_even_partition():
    """That a dataset with even molfluxs is partitioned correctly."""
    train_fraction = 0.6
    validation_fraction = 0.3
    dataset = range(100)
    train_cutoff, validation_cutoff = partition(
        dataset,
        train_fraction,
        validation_fraction,
    )
    assert train_cutoff == 60
    assert validation_cutoff == 90


def test_odd_partition():
    """That a dataset with odd molfluxs is partitioned correctly."""
    train_fraction = 0.6
    validation_fraction = 0.3
    dataset = range(101)
    train_cutoff, validation_cutoff = partition(
        dataset,
        train_fraction,
        validation_fraction,
    )
    assert train_cutoff == 60
    assert validation_cutoff == 90
    # spill over should go into test dataset


def test_train_test_partition():
    """That a train test split gives an empty validation partition."""
    train_fraction = 0.6
    validation_fraction = 0.0
    dataset = range(100)
    train_cutoff, validation_cutoff = partition(
        dataset,
        train_fraction,
        validation_fraction,
    )
    assert train_cutoff == 60
    assert validation_cutoff == 60



import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.uncertainty.uncertainty_based_rejection import (
    UncertaintyBasedRejection,
)


@pytest.fixture()
def fixture_metric():
    return load_metric("uncertainty_based_rejection")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "uncertainty_based_rejection" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, UncertaintyBasedRejection)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_default_compute(fixture_metric):
    """That default scoring."""
    metric = fixture_metric
    predictions = [3, 0.0, 2, 8]
    references = [3, -0.0, 3, 6]
    uncertainties = [0.0, 0.0, 1, 2]
    metric_name = "mean_squared_error"
    result = metric.compute(
        predictions=predictions,
        references=references,
        standard_deviations=uncertainties,
        metric_name=metric_name,
    )

    # The highest threshold should be the same as computing over the entire dataset.
    results_dic = result["uncertainty_based_rejection"]
    threshold_index = max(list(results_dic.keys()))
    hf_metric = load_metric(metric_name)
    result_ind = hf_metric.compute(predictions=predictions, references=references)[
        metric_name
    ]
    assert results_dic[threshold_index][metric_name] == result_ind
    # The entire dataset should be included in the highest threshold.
    assert results_dic[threshold_index]["size_of_thresholded_ds"] == len(predictions)
    assert results_dic[threshold_index]["frac_of_retained_data"] == 1

    # Only the predictions with low uncertainty.
    filtered_pred = predictions[:2]
    filtered_ref = references[:2]
    result_ind = hf_metric.compute(predictions=filtered_pred, references=filtered_ref)[
        metric_name
    ]
    assert results_dic[0][metric_name] == result_ind


def test_different_metric_name(fixture_metric):
    """Change computed kw metric."""
    metric = fixture_metric
    predictions = [3, 0.0, 2, 8]
    references = [3, 0.0, 3, 6]
    uncertainties = [0.0, 0.0, 1, 2]
    metric_name = "mean_absolute_error"
    result = metric.compute(
        predictions=predictions,
        references=references,
        standard_deviations=uncertainties,
        metric_name=metric_name,
    )

    # The highest threshold should be the same as computing over the entire dataset.
    results_dic = result["uncertainty_based_rejection"]
    threshold_index = max(list(results_dic.keys()))
    hf_metric = load_metric(metric_name)
    result_ind = hf_metric.compute(predictions=predictions, references=references)[
        metric_name
    ]
    assert results_dic[threshold_index][metric_name] == result_ind


def test_raise_error_different_length(fixture_metric):
    """That default scoring."""
    metric = fixture_metric
    predictions = [3, 0.0, 2]
    references = [3, -0.0, 3, 1]
    uncertainties = [0.0, 0.0, 1]
    metric_name = "mean_squared_error"
    with pytest.raises(RuntimeError):
        metric.compute(
            predictions=predictions,
            references=references,
            standard_deviations=uncertainties,
            metric_name=metric_name,
        )


def test_num_of_threshold_steps(fixture_metric):
    """That default scoring."""
    metric = fixture_metric
    predictions = [3, 0.0, 2, 8]
    references = [3, -0.0, 3, 6]
    uncertainties = [0.0, 0.0, 1, 2]
    metric_name = "mean_squared_error"
    result = metric.compute(
        predictions=predictions,
        references=references,
        standard_deviations=uncertainties,
        metric_name=metric_name,
        num_of_threshold_steps=2,
    )

    # The number of molfluxs of the dict should be <= num_of_threshold_steps
    # (not necessarily equal since metric will not compute if filtered preds
    # and refs have len < 2
    results_dic = result["uncertainty_based_rejection"]
    assert len(results_dic) <= 2


def test_compute_works_with_prediction_intervals(fixture_metric):
    """Check that using prediction intervals rather than uncertainties directly as input works."""
    metric = fixture_metric
    predictions = [3, 0.0, 2, 8]
    references = [3, -0.0, 3, 6]
    lower_bound = [0.0, 0.0, 1, 2]
    upper_bound = [5.0, 0.9, 4.0, 15.0]
    prediction_intervals = list(zip(lower_bound, upper_bound, strict=False))
    metric_name = "mean_squared_error"
    result = metric.compute(
        predictions=predictions,
        references=references,
        standard_deviations=None,
        prediction_intervals=prediction_intervals,
        metric_name=metric_name,
    )
    assert "uncertainty_based_rejection" in result



from collections.abc import Iterable, Iterator
from typing import Any

import sklearn.model_selection

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable

_DESCRIPTION = """
Stratified K-Folds cross-validator.

Provides train/validation indices to split data in train/validation sets.
This cross-validation object is a variation of KFold that returns
stratified folds. The folds are made by preserving the percentage of
samples for each class.
"""


class StratifiedKFold(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        n_splits: int = 2,
        shuffle: bool = False,
        seed: int | None = None,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            dataset: The data to be split.
            y: The target variable for supervised learning problems.
            groups (optional): Group labels for the samples used while splitting the dataset.
            n_splits (optional): The number of splits to generate. Must be greater than one. Defaults to 2.
            shuffle (optional): Whether to shuffle the data before splitting into batches.
                Note that the samples within each split will not be shuffled.
                Defaults to False.
            seed (optional): When `shuffle` is True, `seed` affects the ordering of the
                indices, which controls the randomness of each fold. Otherwise, this
                parameter has no effect. Pass an int for reproducible output across multiple function calls.
                Defaults to None.
        """
        if y is None:
            raise ValueError("The 'y' parameter should not be None.")

        # Wrap scikit-learn implementation
        cv = sklearn.model_selection.StratifiedKFold(
            n_splits=n_splits,
            shuffle=shuffle,
            random_state=seed,
        )

        for indices in cv.split(dataset, y=y, groups=groups):
            train_indices, validation_indices = indices
            test_indices: Iterable[int] = []  # no holdout in cross-validation
            yield train_indices, validation_indices, test_indices



from pathlib import Path
from typing import Any

import numpy as np
from pydantic.v1 import dataclasses

import datasets
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import (
    ModelBase,
    ModelConfig,
    PredictionIntervalMixin,
    SamplingMixin,
    StandardDeviationMixin,
    UncertaintyCalibrationMixin,
)
from molflux.modelzoo.typing import PredictionResult
from molflux.modelzoo.utils import (
    get_concatenated_array,
    pick_features,
    validate_features,
)

try:
    import stan
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("pystan", e) from None


_DESCRIPTION = """
Implementation of bayesian linear regression with a sparsity inducing prior
See this paper: https://projecteuclid.org/journals/electronic-journal-of-statistics/volume-11/issue-2/Sparsity-information-and-regularization-in-the-horseshoe-and-other-shrinkage/10.1214/17-EJS1337SI.full
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
num_chains : int, default=4  Number of MCMC chains to run. These can be run in parallel allowing faster sampling,
    and having multiple chains is also useful in diagnosing convergence.
num_warmup: int, default=1000 Number of warmup/burnin samples to run to help ensure sampling
    from stationary distribution.
num_samples: int, default=1000 Number of MCMC samples produced per chain. More samples allows for more accurate
    estimation of expectations, and will help ensure convergence, but will take longer to compute.
random_seed: Optional[int], default=None Random seed to be used during sampling for reproducibility
"""


class BayesLinearRegressorHorseshoePrior:
    def __init__(self) -> None:
        path_to_stan_code = (
            Path(__file__).parent / "BayesLinearRegressorHorseshoePrior.stan"
        )
        self.model_code = path_to_stan_code.read_text()
        self.fit = None  # to store posterior samples
        self.beta = None  # to store average coefficients, beta


class Config:
    arbitrary_types_allowed = True


@dataclasses.dataclass(config=Config)
class SparseLinearRegressorConfig(ModelConfig):
    num_chains: int = 4
    num_warmup: int = 1000
    num_samples: int = 1000
    random_seed: int | None = None

    def __post_init_post_parse__(self) -> None:
        if self.y_features and len(self.y_features) != 1:
            raise NotImplementedError(
                f"This model architecture only supports single task regression for now: got {self.y_features}",
            )


class SparseLinearRegressor(
    UncertaintyCalibrationMixin,
    PredictionIntervalMixin,
    StandardDeviationMixin,
    SamplingMixin,
    ModelBase[SparseLinearRegressorConfig],
):
    @property
    def _config_builder(self) -> type[SparseLinearRegressorConfig]:
        return SparseLinearRegressorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> BayesLinearRegressorHorseshoePrior:
        return BayesLinearRegressorHorseshoePrior()

    def _train(self, train_data: datasets.Dataset, **kwargs: Any) -> Any:
        validate_features(train_data, self.y_features)

        x_data = pick_features(train_data, self.x_features)
        X = get_concatenated_array(x_data, self.x_features)

        y_data = pick_features(train_data, self.y_features)
        y = get_concatenated_array(y_data, self.y_features)
        y = y.ravel()

        n, p = X.shape
        data = {"n": n, "p": p, "X": X, "y": y}
        self.model = self._instantiate_model()
        posterior = stan.build(
            program_code=self.model.model_code,
            data=data,
            random_seed=self.model_config.random_seed,
        )
        fit = posterior.sample(
            num_chains=self.model_config.num_chains,
            num_samples=self.model_config.num_samples,
        )
        self.model.fit = fit
        self.model.beta = np.mean(fit["beta"], axis=1)

    def _predict(self, data: datasets.Dataset, **kwargs: Any) -> PredictionResult:
        display_names = self._predict_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        X = get_concatenated_array(data, self.x_features)
        ypred = np.dot(X, self.model.beta)

        return {display_name: ypred.tolist() for display_name in display_names}

    def _predict_with_std(
        self,
        data: datasets.Dataset,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        (
            prediction_display_names,
            prediction_std_display_names,
        ) = self._predict_with_std_display_names

        if not len(data):
            return {display_name: [] for display_name in prediction_display_names}, {
                display_name: [] for display_name in prediction_std_display_names
            }
        X = get_concatenated_array(data, self.x_features)
        ypred = np.dot(X, self.model.beta)
        beta_samples = self.model.fit["beta"]
        sigma_samples = self.model.fit["sigma"]
        mu = np.matmul(X, beta_samples)
        pred = mu + np.random.randn(*sigma_samples.shape) * sigma_samples
        pred_std = np.std(pred, axis=1)

        return {
            display_name: ypred.tolist() for display_name in prediction_display_names
        }, {
            display_name: pred_std.tolist()
            for display_name in prediction_std_display_names
        }

    def _predict_with_prediction_interval(
        self,
        data: datasets.Dataset,
        confidence: float,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        (
            prediction_display_names,
            prediction_interval_display_names,
        ) = self._predict_with_prediction_interval_display_names

        if not len(data):
            return {display_name: [] for display_name in prediction_display_names}, {
                display_name: [] for display_name in prediction_interval_display_names
            }

        X = get_concatenated_array(data, self.x_features)
        ypred = np.dot(X, self.model.beta)
        beta_samples = self.model.fit["beta"]
        sigma_samples = self.model.fit["sigma"]
        mu = np.matmul(X, beta_samples)
        pred = mu + np.random.randn(*sigma_samples.shape) * sigma_samples
        lower_bound = np.quantile(pred, (1 - confidence) / 2.0, axis=1)
        upper_bound = np.quantile(pred, 1 - (1 - confidence) / 2.0, axis=1)
        # where nans occur (ex. from a 0 standard deviation), use the mean instead
        lower_bound = np.where(~np.isnan(lower_bound), lower_bound, ypred)
        upper_bound = np.where(~np.isnan(upper_bound), upper_bound, ypred)

        return {
            display_name: ypred.tolist() for display_name in prediction_display_names
        }, {
            display_name: list(
                zip(lower_bound.tolist(), upper_bound.tolist(), strict=False),
            )
            for display_name in prediction_interval_display_names
        }

    def _sample(
        self,
        data: datasets.Dataset,
        n_samples: int,
        **kwargs: Any,
    ) -> PredictionResult:
        display_names = self._sample_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        prediction_mean_results, prediction_std_results = self._predict_with_std(data)

        prediction_results: PredictionResult = {}
        for display_name, means, stds in zip(
            display_names,
            prediction_mean_results.values(),
            prediction_std_results.values(),
            strict=False,
        ):
            samples = np.random.normal(means, stds, (n_samples, len(means))).T
            prediction_results[display_name] = samples.tolist()

        return prediction_results

    def _calibrate_uncertainty(self, data: datasets.Dataset, **kwargs: Any) -> Any:
        pass



if __name__ == "__main__":
    from pathlib import Path

    import h5py

    import datasets
    from molflux.datasets.builders.ani2x.ani2x_configs import FEATURES, URL_DICT

    dl_manager = datasets.DownloadManager()
    Path("final_h5").mkdir()

    for level_of_theory in URL_DICT.keys():
        h5_file_path = dl_manager.download_and_extract(URL_DICT[level_of_theory]["url"])
        h5_file = h5py.File(f"{h5_file_path}/{URL_DICT[level_of_theory]['path']}", "r")

        new_h5_file = h5py.File(URL_DICT[level_of_theory]["path"], "w")
        for key in list(h5_file.keys())[:5]:
            new_h5_file.create_group(key)
            for feature in FEATURES[level_of_theory]:
                if feature != "mol_bytes":
                    new_h5_file[key].create_dataset(
                        feature,
                        data=h5_file[key][feature][:2],
                    )
            new_h5_file[key].create_dataset(
                "species",
                data=h5_file[key]["species"][:2],
            )
            new_h5_file[key].create_dataset(
                "coordinates",
                data=h5_file[key]["coordinates"][:2],
            )

        new_h5_file.close()



import os
import uuid
from pathlib import Path
from types import SimpleNamespace

import fsspec
import pandas as pd
import pytest
from cloudpathlib import S3Path

import datasets
from molflux.datasets import (
    load_dataset,
    load_dataset_from_store,
    save_dataset_to_store,
)


@pytest.fixture(
    params=[
        # simple path
        "data.parquet",
        # path with potentially problematic characters
        "2023-03-22T17:53:17.245386.parquet",
    ],
)
def fixture_sample_parquet_dataset(tmpdir_factory, request):
    """A temporary sample parquet dataset."""
    data = {"col1": [1, 2], "col2": [3, 4]}
    df = pd.DataFrame(data=data)

    filename = request.param
    path = str(tmpdir_factory.mktemp("data").join(filename))
    df.to_parquet(path, index=False)

    shape = df.shape
    columns = list(data.keys())
    return SimpleNamespace(path=path, shape=shape, columns=columns)


@pytest.fixture(scope="module")
def fixture_collection_of_parquet_datasets(tmpdir_factory):
    """A couple of temporary sample parquet dataset."""
    data = {"col1": [1, 2], "col2": [3, 4]}
    df = pd.DataFrame(data=data)

    filename_one = str(tmpdir_factory.mktemp("data").join("data_one.parquet"))
    df.to_parquet(filename_one, index=False)

    filename_two = str(tmpdir_factory.mktemp("data").join("data_one.parquet"))
    df.to_parquet(filename_two, index=False)

    return filename_one, filename_two


@pytest.fixture(scope="module")
def fixture_sample_parquet_dataset_dict(tmpdir_factory):
    """A temporary sample DatasetDict saved as parquet."""

    data = datasets.Dataset.from_dict({"col1": [1, 2], "col2": [3, 4]})
    dataset_dict = datasets.DatasetDict(
        {
            "train": data,
            "validation": data,
            "test": data,
        },
    )

    dir_name = str(tmpdir_factory.mktemp("data").join("data"))
    save_dataset_to_store(dataset_dict, path=dir_name, format="parquet")
    return dir_name


def test_load_from_stored_parquet(fixture_sample_parquet_dataset):
    """That can load a parquet file as a Dataset."""
    parquet = fixture_sample_parquet_dataset

    # test both against str-paths and Path-like paths
    str_path = str(parquet.path)
    pathlib_path = Path(parquet.path)

    for path in [str_path, pathlib_path]:
        dataset = load_dataset_from_store(path, format="parquet")  # type: ignore[arg-type]
        assert isinstance(dataset, datasets.Dataset)


def test_can_load_from_stored_parquet_without_specifying_format(
    fixture_sample_parquet_dataset,
):
    """That can load a single parquet file as a Dataset without explicitly
    specifying the data 'format'."""
    parquet = fixture_sample_parquet_dataset
    dataset = load_dataset_from_store(parquet.path)
    assert isinstance(dataset, datasets.Dataset)


def test_can_load_from_formatless_file_if_format_provided_explicitly(tmp_path):
    """That explicitly specifying a file format allows loading datasets that
    whose format would not otherwise be automatically resolvable.
    """
    data = {"col1": [1, 2], "col2": [3, 4]}
    df = pd.DataFrame(data=data)
    path = str(tmp_path / "data")
    df.to_parquet(path, index=False)

    dataset = load_dataset_from_store(path, format="parquet")
    assert isinstance(dataset, datasets.Dataset)


def test_load_from_sequence_of_files(fixture_collection_of_parquet_datasets):
    """That can load a sequence of (coherent) files as a single Dataset."""
    path_one, path_two = fixture_collection_of_parquet_datasets

    # test both against str-paths and Path-like paths
    str_data_files = [str(path_one), str(path_two)]
    pathlib_data_files = [Path(path_one), Path(path_two)]

    for data_files in [str_data_files, pathlib_data_files]:
        dataset = load_dataset_from_store(data_files, format="parquet")  # type: ignore[arg-type]
        assert isinstance(dataset, datasets.Dataset)


def test_load_from_mapping_of_files_as_dataset_dict(
    fixture_collection_of_parquet_datasets,
):
    """That can load a mapping of (coherent) files as a single DatasetDict."""
    path_one, path_two = fixture_collection_of_parquet_datasets

    # test against str-paths and Path-like paths
    str_mapping = {
        "train": str(path_one),
        "test": str(path_two),
    }
    pathlib_mapping = {
        "train": Path(path_one),
        "test": Path(path_two),
    }

    for data_files in [str_mapping, pathlib_mapping]:
        dataset = load_dataset_from_store(data_files, format="parquet")  # type: ignore[arg-type]
        assert isinstance(dataset, datasets.DatasetDict)

        dataset = load_dataset_from_store(data_files, format="parquet", split=None)  # type: ignore[arg-type]
        assert isinstance(dataset, datasets.DatasetDict)


def test_load_from_mapping_of_files_as_dataset(
    fixture_collection_of_parquet_datasets,
):
    """That can load a mapping of (coherent) files as a single Dataset."""
    path_one, path_two = fixture_collection_of_parquet_datasets
    mapping = {
        "train": path_one,
        "test": path_two,
    }
    dataset = load_dataset_from_store(mapping, format="parquet", split="all")
    assert isinstance(dataset, datasets.Dataset)

    dataset = load_dataset_from_store(mapping, format="parquet", split="train")
    assert isinstance(dataset, datasets.Dataset)

    dataset = load_dataset_from_store(mapping, format="parquet", split="test")
    assert isinstance(dataset, datasets.Dataset)


def test_load_from_dataset_dict_dir(fixture_sample_parquet_dataset_dict):
    """That can load a persisted DatasetDict."""
    path = fixture_sample_parquet_dataset_dict

    dataset_dict = load_dataset_from_store(path, format="parquet", split=None)
    assert isinstance(dataset_dict, datasets.DatasetDict)

    # Test default returns None splits (i.e. a DatasetDict)
    dataset_dict = load_dataset_from_store(path, format="parquet")
    assert isinstance(dataset_dict, datasets.DatasetDict)

    # Test can load a specific split
    dataset = load_dataset_from_store(path, format="parquet", split="train")
    assert isinstance(dataset, datasets.Dataset)

    # Test can load all splits as single Dataset
    dataset = load_dataset_from_store(path, format="parquet", split="all")
    assert isinstance(dataset, datasets.Dataset)


def test_load_from_dataset_dict_dir_defaults_to_assume_parquet(
    fixture_sample_parquet_dataset_dict,
):
    """That loading a DatasetDict defaults to assuming data has been persisted
    as parquet files."""
    path = fixture_sample_parquet_dataset_dict

    dataset_dict = load_dataset_from_store(path, split=None)
    assert isinstance(dataset_dict, datasets.DatasetDict)


def test_load_from_cloud_parquet(
    fixture_mock_s3_client,
    fixture_mock_s3_filesystem,
    fixture_sample_parquet_dataset,
):
    """That can load a dataset saved as parquet from AWS S3."""

    # Upload the dataset to the mock S3 to simulate having it already there
    parquet = fixture_sample_parquet_dataset
    mock_bucket_name = str(uuid.uuid4())
    client = fixture_mock_s3_client
    client.create_bucket(Bucket=mock_bucket_name)
    client.upload_file(
        Filename=parquet.path,
        Bucket=mock_bucket_name,
        Key="data.parquet",
    )

    # Load dataset

    # test both against str-paths and Path-like paths
    str_path = f"s3://{mock_bucket_name}/data.parquet"
    pathlib_path = S3Path(str_path)

    for path in [str_path, pathlib_path]:
        fs = fixture_mock_s3_filesystem
        dataset = load_dataset_from_store(path, format="parquet", fs=fs)  # type: ignore[arg-type]
        assert isinstance(dataset, datasets.Dataset)


@pytest.fixture(scope="module")
def fixture_dataset_saved_to_disk(tmpdir_factory):
    path = str(tmpdir_factory.mktemp("fixture_dataset"))
    dataset = load_dataset("esol")
    fs = fsspec.filesystem("file")
    save_dataset_to_store(dataset, path=path, format="disk", fs=fs)
    return SimpleNamespace(path=path, shape=dataset.shape, columns=dataset.column_names)


def test_load_from_stored_disk_datadir(fixture_dataset_saved_to_disk):
    """That can load a dataset saved as 'disk' as a Dataset."""
    disk = fixture_dataset_saved_to_disk
    dataset = load_dataset_from_store(disk.path, format="disk")
    assert isinstance(dataset, datasets.Dataset)


def test_load_from_cloud_disk_datadir(
    fixture_mock_s3_client,
    fixture_mock_s3_filesystem,
    fixture_dataset_saved_to_disk,
):
    """That can load a dataset saved as 'disk' from AWS S3."""

    # Upload the dataset to the mock S3 to simulate having it already there
    saved_dataset = fixture_dataset_saved_to_disk
    mock_bucket_name = str(uuid.uuid4())
    client = fixture_mock_s3_client
    client.create_bucket(Bucket=mock_bucket_name)
    for f in os.listdir(saved_dataset.path):
        client.upload_file(
            Filename=os.path.join(saved_dataset.path, f),
            Bucket=mock_bucket_name,
            Key=f"data/{f}",
        )

    # test both against str-paths and Path-like paths
    str_path = f"s3://{mock_bucket_name}/data"
    pathlib_path = S3Path(str_path)

    for path in [str_path, pathlib_path]:
        fs = fixture_mock_s3_filesystem
        dataset = load_dataset_from_store(path, format="disk", fs=fs)  # type: ignore[arg-type]
        assert isinstance(dataset, datasets.Dataset)



import numpy.testing
import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.classification.validity_roc import ValidityRoc

metric_name = "validity_roc"


@pytest.fixture()
def fixture_metric():
    return load_metric(metric_name)


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert metric_name in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, ValidityRoc)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_string_labels(fixture_metric):
    """That can use string labels."""
    metric = fixture_metric
    predictions = [["CC", "CC", "C", "C", "CCC"]]
    references = [None]
    result = metric.compute(predictions=predictions, references=references)
    numpy.testing.assert_allclose([[1.0, 1.0, 1.0, 1.0, 1.0]], result[metric_name])


def test_none_labels(fixture_metric):
    """That can use None labels."""
    metric = fixture_metric
    predictions = [["A", "A", None, None, "B"]]
    references = [None]
    result = metric.compute(predictions=predictions, references=references)
    numpy.testing.assert_allclose([[1.0, 1.0, 2 / 3, 0.5, 0.6]], result[metric_name])


def test_custom_invalid_token(fixture_metric):
    """That can use custom invalid tokens."""
    metric = fixture_metric
    predictions = [["CC", "CC", "<INVALID>", "<INVALID>", "CCC"]]
    references = [None]
    result = metric.compute(
        predictions=predictions,
        references=references,
        invalid_token="<INVALID>",  # noqa: S106
    )
    numpy.testing.assert_allclose([[1.0, 1.0, 2 / 3, 0.5, 0.6]], result[metric_name])


def test_batch_validity(fixture_metric):
    """That metric can be calculated on list of lists."""
    metric = fixture_metric
    predictions = [["A", "B", "B", None, "C"], ["D", "D", "D", "E", "F"]]
    references = [None, None]
    result = metric.compute(predictions=predictions, references=references)
    numpy.testing.assert_allclose(
        [[1.0, 1.0, 1.0, 0.75, 0.8], [1.0, 1.0, 1.0, 1.0, 1.0]],
        result[metric_name],
    )


def test_with_average(fixture_metric):
    """That switching on the average flag returns expected results."""
    metric = fixture_metric
    predictions = [["A", "B", "B", None, "C"], ["D", "D", "D", "E", "F"]]
    references = [None, None]
    result = metric.compute(
        predictions=predictions,
        references=references,
        average=True,
    )
    numpy.testing.assert_allclose([[1.0, 1.0, 1.0, 0.875, 0.9]], result[metric_name])



"""Prediction interval width for models with interval predictions"""

import logging
from typing import Any

import evaluate
import numpy as np
import scipy

import datasets
from molflux.metrics.bases import UncertaintyMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Prediction Interval Width function.

Predictions for models with uncertainty may provide lower and upper bounds for
predictions at a certain confidence level. Assessing the width of these intervals
across a dataset can help assess how sharp these predictions are.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: array of mean predictions.
    references: Ground truth (correct) target values.
    lower_bound: array of lower bounds of predictions.
    upper_bound: array of upper bounds of predictions.

Returns:
    prediction_interval_width: The value of the prediction interval width.

Examples:
    >>> from molflux.metrics import load_metric
    >>> m = load_metric("prediction_interval_width")
    >>> ref = [0, 1, 2, 3, 4]
    >>> pred = [0, 2, 2, 3, 5]
    >>> lower_bound = [0.1, 0.1, 0.2, 2, 3]
    >>> upper_bound = [1, 1.5, 1.8, 2.8, 5.7]
    >>> prediction_intervals = list(zip(lower_bound, upper_bound))
    >>> m.compute(predictions=pred, references=ref, prediction_intervals=prediction_intervals)
    {'prediction_interval_width': 1.48}
"""

_CITATION = """\
None
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class PredictionIntervalWidth(UncertaintyMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float64"),
                    "references": datasets.Value("float64"),
                },
            ),
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        standard_deviations: ArrayLike | None = None,
        prediction_intervals: ArrayLike | None = None,
        confidence: float = 0.9,
        **kwargs: Any,
    ) -> MetricResult:
        if standard_deviations is not None and prediction_intervals is not None:
            raise ValueError(
                "Both standard deviations and prediction intervals given. Please provide only one.",
            )

        if standard_deviations is not None:
            lower_bound, upper_bound = scipy.stats.norm.interval(
                confidence,
                loc=predictions,
                scale=standard_deviations,
            )
        elif prediction_intervals is not None:
            lower_bound, upper_bound = zip(*prediction_intervals, strict=False)
        else:
            raise ValueError(
                "Either standard_deviations or prediction_intervals must be supplied.",
            )

        # ensure the predictions are a numpy array
        lower_bound = np.array(lower_bound)
        upper_bound = np.array(upper_bound)
        if (upper_bound < lower_bound).any():
            raise ValueError("Please ensure upper bound is greater than lower bound.")
        width = np.mean(upper_bound - lower_bound)

        return {self.tag: width}



import numbers
from typing import Any

import numpy as np

from datasets import Dataset

MAX_INT = np.iinfo(np.int32).max
MIN_INT = -1 * MAX_INT - 1


def check_parameter(
    param: int | float,
    low: int | float = MIN_INT,
    high: int | float = MAX_INT,
    param_name: str = "",
    include_left: bool = False,
    include_right: bool = False,
) -> bool:
    """Check if an input is within the defined range.
    Parameters
    ----------
    param : int, float
        The input parameter to check.
    low : int, float
        The lower bound of the range.
    high : int, float
        The higher bound of the range.
    param_name : str, optional (default='')
        The name of the parameter.
    include_left : bool, optional (default=False)
        Whether includes the lower bound (lower bound <=).
    include_right : bool, optional (default=False)
        Whether includes the higher bound (<= higher bound).
    Returns
    -------
    within_range : bool or raise errors
        Whether the parameter is within the range of (low, high)
    """

    # param, low and high should all be numerical
    if not isinstance(param, (numbers.Integral, np.integer, float)):
        raise TypeError(f"{param_name} is set to {param} Not numerical")

    if not isinstance(low, (numbers.Integral, np.integer, float)):
        raise TypeError(f"low is set to {low}. Not numerical")

    if not isinstance(high, (numbers.Integral, np.integer, float)):
        raise TypeError(f"high is set to {high}. Not numerical")

    # at least one of the bounds should be specified
    if low is MIN_INT and high is MAX_INT:
        raise ValueError("Neither low nor high bounds is undefined")

    # if wrong bound values are used
    if low > high:
        raise ValueError("Lower bound > Higher bound")

    # value check under different bound conditions
    if (include_left and include_right) and (param < low or param > high):
        raise ValueError(
            f"{param_name} is set to {param}. " f"Not in the range of [{low}, {high}].",
        )

    elif (include_left and not include_right) and (param < low or param >= high):
        raise ValueError(
            f"{param_name} is set to {param}. " f"Not in the range of [{low}, {high}).",
        )

    elif (not include_left and include_right) and (param <= low or param > high):
        raise ValueError(
            f"{param_name} is set to {param}. " f"Not in the range of ({low}, {high}].",
        )

    elif (not include_left and not include_right) and (param <= low or param >= high):
        raise ValueError(
            f"{param_name} is set to {param}. " f"Not in the range of ({low}, {high}).",
        )
    else:
        return True


def list_diff(first_list: list, second_list: list) -> list | set:
    """Utility function to calculate list difference (first_list-second_list)
    For duplicated values, both duplicates are removed such that
    eg. >>> list_diff([1,2,2],[1,2])
        []
    Parameters
    ----------
    first_list : list
        First list.
    second_list : list
        Second list.
    Returns
    -------
    diff : different molfluxs.
    """
    second_set = set(second_list)
    return [item for item in first_list if item not in second_set]


def get_split_indices(
    dataset: Dataset,
    n_folds: int = 3,
) -> Any:
    """Utility function to get split indices to then splits the data
    for stacking. The data is split
    into n_folds with roughly equal rough size.
    Parameters
    ----------
    dataset: Dataset
    n_folds : int, optional (default=3)
        The number of splits of the training sample.
    Returns
    -------
    index_lists : list of list
        The list of indexes of each fold regarding the returned X and y.
        For instance, index_lists[0] contains the indexes of fold 0.
    """

    if not isinstance(n_folds, int):
        raise ValueError("n_folds must be an integer variable")
    check_parameter(n_folds, low=2, include_left=True, param_name="n_folds")

    idx_length = dataset.shape[0]
    idx_list = list(range(idx_length))

    avg_length = int(idx_length / n_folds)

    index_lists = []
    for i in range(n_folds - 1):
        index_lists.append(idx_list[i * avg_length : (i + 1) * avg_length])

    index_lists.append(idx_list[(n_folds - 1) * avg_length :])

    return index_lists



import json
import os
from dataclasses import asdict, dataclass, field
from typing import Any

from molflux.features import config


@dataclass
class RepresentationInfo:
    """
    Information about a representation.

    See the constructor arguments and properties for a full list.

    Note: Not all fields are known on construction and may be updated later.
    """

    # required
    description: str

    # Optional
    version: str | None = None

    # set later by the builder / initialiser
    name: str = field(default_factory=str)
    tag: str = field(default_factory=str)
    featurise_description: str = field(default_factory=str)

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)

    def write_to_directory(self, directory: str) -> None:
        log_file = os.path.join(directory, config.REPRESENTATION_INFO_FILENAME)
        with open(log_file, "w", encoding="utf-8") as f:
            json.dump(self.to_dict(), f, indent=4)



import pytest

from molflux.metrics.load import load_metric


@pytest.fixture(scope="module")
def fixture_mock_metric():
    return load_metric(kind="classification", name="accuracy")


def test_invalid_kwarg_raises_value_error(fixture_mock_metric):
    """That attempting scoring with invalid keyword arguments raises."""
    metric = fixture_mock_metric
    predictions = [1, 0]
    references = [1, 0]
    with pytest.raises(ValueError, match=r"Unknown compute parameter\(s\)"):
        metric.compute(
            predictions=predictions,
            references=references,
            invalid_kwarg=True,
        )



import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.classification.top_k_accuracy import TopKAccuracy

metric_name = "top_k_accuracy"


@pytest.fixture()
def fixture_metric():
    return load_metric(metric_name)


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert metric_name in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, TopKAccuracy)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_multiclass(fixture_metric):
    """Multiclass case."""
    metric = fixture_metric
    predictions = [
        [0.5, 0.2, 0.2],  # 0 is in top 2
        [0.3, 0.4, 0.2],  # 1 is in top 2
        [0.2, 0.4, 0.3],  # 2 is in top 2
        [0.7, 0.2, 0.1],  # 2 isn't in top 2
    ]
    references = [0, 1, 2, 2]
    result = metric.compute(predictions=predictions, references=references)
    assert result["top_k_accuracy"] == 3 / 4



from unittest.mock import patch

import pytest

import datasets
from molflux.datasets import load_dataset
from molflux.datasets.builders.esol.esol import ESOL
from molflux.datasets.catalogue import list_datasets

dataset_name = "esol"


@pytest.fixture()
def _fixture_import_main_class_patch():
    """Mocks out the import_main_class from datasets'.

    This function is used when generating a dataset builder in the `datasets` class.
    It is done by copying the file path of a builder script into a temporary folder
    under the huggingface cache directory, and loading the Builder class dynamically
    at runtime. This means that any test for a builder will instead be tested on a
    copy of the builder script, not showing up as covered in any coverage run.
    Instead, we redirect the builder class import to the actual class in alexandria.
    """
    with patch("datasets.load.import_main_class") as mock:
        mock.return_value = ESOL
        yield


pytestmark = pytest.mark.usefixtures("_fixture_import_main_class_patch")


def test_in_catalogue() -> None:
    """That the builder is registered in the catalogue."""
    catalogue = list_datasets()
    all_names = [name for names in catalogue.values() for name in names]
    assert dataset_name in all_names


def test_returns_huggingface_dataset():
    """That the built dataset is a huggingface dataset."""
    dataset = load_dataset(dataset_name, split="train")
    assert isinstance(dataset, datasets.Dataset)


def test_default_config_is_available():
    """That the dataset can be loaded without specifying an explicit config."""
    assert load_dataset(dataset_name)


def test_defines_expected_feature_names():
    """That the dataset defines these feature names.

    This test makes sure that we can catch possibly breaking changes to consumers
    if we rename these features down the line.
    """
    expected_feature_names = [
        "smiles",
        "log_solubility",
    ]
    dataset = load_dataset(dataset_name)

    feature_names = dataset.column_names
    assert feature_names is not None
    assert all(name in feature_names for name in expected_feature_names)
    assert len(feature_names) == len(expected_feature_names)



import logging
import os
from abc import abstractmethod
from dataclasses import asdict, is_dataclass, replace
from typing import (
    Any,
)

from datasets import Dataset
from molflux.modelzoo.errors import NotTrainedError
from molflux.modelzoo.model import ModelBase
from molflux.modelzoo.typing import PredictionResult

try:
    import lightning.pytorch as pl
    import torch
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning", e) from e

from molflux.modelzoo.models.lightning.config import (
    CompileConfig,
    DataModuleConfig,
    LightningConfigT,
    OptimizerConfig,
    SchedulerConfig,
    TrainerConfig,
    TransferLearningConfigBase,
)
from molflux.modelzoo.models.lightning.datamodule import LightningDataModule
from molflux.modelzoo.models.lightning.module import LightningModuleBase
from molflux.modelzoo.models.lightning.utils import load_from_dvc_or_disk

logger = logging.getLogger(__name__)


class LightningModelBase(ModelBase[LightningConfigT]):
    """ABC for all PyTorch Lightning"""

    _train_multi_data_enabled: bool = True

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        precision = self.model_config.float32_matmul_precision
        if precision is not None:
            torch.set_float32_matmul_precision(precision)

    @property
    @abstractmethod
    def _datamodule_builder(self) -> type[LightningDataModule]:
        """The DataModule for this model class.

        Implementations should subclass from the ModelZoo LightningDataModule."""
        ...

    def _instantiate_datamodule(
        self,
        train_data: dict[str | None, Dataset] | None = None,
        validation_data: dict[str | None, Dataset] | None = None,
        test_data: dict[str | None, Dataset] | None = None,
        predict_data: Dataset | None = None,
        **kwargs: Any,
    ) -> LightningDataModule:
        """Prepares the datamodule.

        Implementations do not have to modify this method."""
        return self._datamodule_builder(
            self.model_config,
            train_data=train_data,
            validation_data=validation_data,
            test_data=test_data,
            predict_data=predict_data,
            **kwargs,
        )

    def _train(self, train_data: Dataset, **kwargs: Any) -> Any:
        """Training from single data sources is disabled for Lightning models."""
        del train_data, kwargs
        raise NotImplementedError("This class implements _train_multi_data instead.")

    def _train_multi_data(
        self,
        train_data: dict[str | None, Dataset],
        validation_data: dict[str | None, Dataset] | None = None,
        datamodule_config: DataModuleConfig | dict[str, Any] | None = None,
        trainer_config: TrainerConfig | dict[str, Any] | None = None,
        optimizer_config: OptimizerConfig | dict[str, Any] | None = None,
        scheduler_config: SchedulerConfig | dict[str, Any] | None = None,
        transfer_learning_config: TransferLearningConfigBase
        | dict[str, Any]
        | None = None,
        compile_config: CompileConfig | dict[str, Any] | bool | None = None,
        ckpt_path: str | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Training method. Lightning models do not need to modify this method.

        Args:
            train_data: Dictionary of training datasets.
            validation_data: Optional validation dataset(s).
            datamodule_config: Optional dict with kwargs to temporarily override the datamodule config in the model config.
            trainer_config: Optional dict with kwargs to temporarily override the trainer config in the model config.
            optimizer_config: Optional dict with kwargs to temporarily override the optimizer config in the model config.
            scheduler_config: Optional dict with kwargs to temporarily override the scheduler config in the model config.
            transfer_learning_config: Optional dict with kwargs to temporarily override the transfer_learning_config config in the model config.
            compile_config: Optional dict with kwargs to temporarily override the compile_config config in the model config.
            ckpt_path: Optional path (local or s3) to a lightning checkpoint to resume training from. Used by the lighting trainer ``.fit`` method.
        Returns:
            None

        """
        del kwargs

        with self.override_config(
            datamodule=datamodule_config,
            trainer=trainer_config,
            transfer_learning=transfer_learning_config,
            optimizer=optimizer_config,
            scheduler=scheduler_config,
            compile=compile_config,
        ):
            if self.model_config.transfer_learning is not None:
                self._transfer_learn(train_data, validation_data)
            else:
                datamodule = self._instantiate_datamodule(
                    train_data=train_data,
                    validation_data=validation_data,
                )
                self.module: Any = self._instantiate_module()
                self._compile()
                trainer = pl.Trainer(**self.model_config.trainer.pass_to_trainer())
                trainer.fit(self.module, datamodule=datamodule, ckpt_path=ckpt_path)

    def _match_modules(
        self,
        new_module: LightningModuleBase,
        old_module: LightningModuleBase,
        modules_to_match: dict[str, str] | None,
    ) -> None:
        if modules_to_match is not None:
            list_of_old_modules = [name for name, _ in old_module.named_modules()]
            list_of_new_modules = [name for name, _ in new_module.named_modules()]

            if not (set(modules_to_match.keys()) <= set(list_of_new_modules)):
                raise KeyError(
                    f"Modules {set(modules_to_match.keys()) - set(list_of_new_modules)} not in new module.",
                )

            if not (set(modules_to_match.values()) <= set(list_of_old_modules)):
                raise KeyError(
                    f"Modules {set(modules_to_match.values()) - set(list_of_old_modules)} not in old module.",
                )

            # iterate over new module's named modules and try to match
            for name, sub_module in new_module.named_modules():
                if name in modules_to_match:
                    try:
                        sub_module.load_state_dict(
                            old_module.get_submodule(
                                modules_to_match[name],
                            ).state_dict(),
                        )
                        logger.warning(
                            f"Matched module '{name}' in new module to '{modules_to_match[name]}' in old module.\n",
                        )
                    except RuntimeError as exc:
                        raise KeyError(
                            f"Unable to match module '{name}' in new module to '{modules_to_match[name]}' in old module. {exc}",
                        ) from exc
        else:
            # try to match entire module
            try:
                new_module.load_state_dict(old_module.state_dict())
                logger.warning("Matched all of old module to new module")
            except RuntimeError as exc:
                raise KeyError(
                    f"Unable to match all of old module to new module. {exc}",
                ) from exc

    def _freeze_modules(
        self,
        module: LightningModuleBase,
        freeze_modules: list[str] | None,
    ) -> None:
        if freeze_modules is not None:
            list_of_modules = [name for name, _ in module.named_modules()]
            if not (set(freeze_modules) <= set(list_of_modules)):
                raise KeyError(
                    f"Modules {set(freeze_modules) - set(list_of_modules)} not in module.",
                )

            for name, sub_module in module.named_modules():
                if name in freeze_modules:
                    sub_module.requires_grad_(False)
                    logger.warning(f"Freezing module: {name}.\n")

    def _transfer_learn(
        self,
        train_data: dict[str | None, Dataset],
        validation_data: dict[str | None, Dataset] | None = None,
    ) -> None:
        if self.model_config.transfer_learning is None:
            return None

        # load pretrained model
        pre_trained_model = load_from_dvc_or_disk(
            path=self.model_config.transfer_learning.pre_trained_model_path,
            repo_url=self.model_config.transfer_learning.repo_url,
            rev=self.model_config.transfer_learning.rev,
            model_path_in_repo=self.model_config.transfer_learning.model_path_in_repo,
        )

        if not isinstance(pre_trained_model, LightningModelBase):
            raise RuntimeError("Pre-trained model is not a LightningModelBase.")

        self.module = self._instantiate_module()
        # match modules
        self._match_modules(
            new_module=self.module,
            old_module=pre_trained_model.module,
            modules_to_match=self.model_config.transfer_learning.modules_to_match,
        )
        ckpt_for_next_stage: dict = self.state_dict

        for stage in self.model_config.transfer_learning.stages:
            with self.override_config(
                trainer=stage.trainer,
                datamodule=stage.datamodule,
                optimizer=stage.optimizer,
                scheduler=stage.scheduler,
            ):
                datamodule = self._instantiate_datamodule(
                    train_data=train_data,
                    validation_data=validation_data,
                )

                self.module = self._instantiate_module()
                self.state_dict = ckpt_for_next_stage

                self._freeze_modules(
                    module=self.module,
                    freeze_modules=stage.freeze_modules,
                )

                self._compile()

                trainer = pl.Trainer(**self.model_config.trainer.pass_to_trainer())
                trainer.fit(self.module, datamodule=datamodule)

                # get ckpt for next stage
                ckpt_for_next_stage = self.state_dict

    def _predict(
        self,
        data: Dataset,
        datamodule_config: DataModuleConfig | dict[str, Any] | None = None,
        trainer_config: TrainerConfig | dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> PredictionResult:
        """Implements a single prediction step.

        Lightning models do not need to modify this step."""
        del kwargs

        display_names = self._predict_display_names

        # if data is empty
        if not len(data):
            return {display_name: [] for display_name in display_names}

        with self.override_config(datamodule=datamodule_config, trainer=trainer_config):
            datamodule = self._instantiate_datamodule(predict_data=data)

            trainer_config = self.model_config.trainer.pass_to_trainer()
            trainer = pl.Trainer(
                accelerator=trainer_config["accelerator"],
                devices=trainer_config["devices"],
                strategy=trainer_config["strategy"],
                logger=False,
            )

            # Expect a list of Tensors, which may need to be overwritten for some Torch models
            batch_preds: list[torch.Tensor] = trainer.predict(  # pyright: ignore
                self.module,
                datamodule,
            )

        batched_preds_catted = torch.cat(batch_preds, dim=0)

        return {
            display_name: batched_preds_catted[:, idx].tolist()
            for idx, display_name in enumerate(display_names)
        }

    @abstractmethod
    def _instantiate_module(self) -> LightningModuleBase:
        """The LightningModule attached to this model class.

        Should subclass the ModelZoo LightningModuleBase"""
        ...

    def _compile(self) -> None:
        """Compiles the Lightning model."""
        compile_config = self.model_config.compile
        if isinstance(compile_config, CompileConfig):
            raise ValueError(
                "Compilation temporarily disabled. It will be re-enabled with PyTorch 2.1.\n"
                "https://github.com/Lightning-AI/lightning/issues/17177#issuecomment-1481105838",
            )
            # TODO bring back with PyTorch 2.1
            # self.module = torch.compile(
            #     self.module,
            #     mode=compile_config.mode,
            #     dynamic=compile_config.dynamic,
            #     fullgraph=compile_config.fullgraph,
            #     backend=compile_config.backend,
            #     **compile_config.backend_kwargs,
            # )

    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    def override_config(self, **section_overrides: Any) -> "ConfigOverride":
        """Context manager for temporarily overriding config, e.g. during
        transfer learning."""
        return ConfigOverride(self, **section_overrides)

    @property
    def is_compiled(self) -> bool:
        return hasattr(self, "module") and hasattr(self.module, "_orig_mod")

    @property
    def state_dict(self) -> Any:
        module = self.module if not self.is_compiled else self.module._orig_mod
        return module.state_dict()

    @state_dict.setter
    def state_dict(self, state_dict: dict[str, torch.Tensor]) -> None:
        module = self.module if not self.is_compiled else self.module._orig_mod
        module.load_state_dict(state_dict)

    def as_dir(self, directory: str) -> None:
        """Saves the model into a directory."""

        if hasattr(self, "module") and self.module is not None:
            ckpt = {"state_dict": self.state_dict}
            torch.save(ckpt, os.path.join(directory, "module_checkpoint.ckpt"))
        else:
            raise NotTrainedError("No initialised module to save.")

    def from_dir(self, directory: str) -> None:
        """Loads a model from a directory."""

        ckpt_path = os.path.join(directory, "module_checkpoint.ckpt")
        checkpoint = torch.load(ckpt_path, map_location="cpu", weights_only=True)

        self.module = self._instantiate_module()
        self.state_dict = checkpoint["state_dict"]


class ConfigOverride:
    def __init__(
        self,
        model: LightningModelBase[LightningConfigT],
        **section_overrides: dict[str, Any] | bool,
    ):
        self.model = model
        self.section_overrides = {
            section: override
            for section, override in section_overrides.items()
            if override is not None
        }
        self.original_config = model.model_config

    def __enter__(self) -> None:
        new_sections = {}
        for section, override in self.section_overrides.items():
            old_section = getattr(self.original_config, section)

            # Checks if override is a dataclass instance (not type) with suitable type
            override_obj: bool | dict[str, Any]
            if (
                is_dataclass(override)
                and not isinstance(override, type)  # type: ignore[unreachable]
                and (
                    isinstance(override, type(old_section))
                    or (old_section is None)
                    or isinstance(old_section, bool)
                )
            ):
                override_obj = asdict(override)  # type: ignore[unreachable]
            elif isinstance(override, dict) or isinstance(override, bool):
                override_obj = override
            else:
                raise RuntimeError(
                    f"Unsure how to override {type(old_section)} with object of type {type(override)}",
                )

            if (
                old_section is not None
                and not isinstance(old_section, bool)
                and not isinstance(override_obj, bool)
            ):
                new_sections[section] = replace(old_section, **override_obj)
            else:
                new_sections[section] = override_obj

        # Mypy has trouble recognising TypeVar bound to a Pydantic dataclass
        new_model_config = replace(
            self.original_config,
            **new_sections,
        )

        # NOTE model_config is stored in the model, module and datamodule
        # It will be correctly overridden ONLY in these two locations.
        # Datamodules are always instantiated UNDER the override and so are
        # passed an overridden config.
        self.model.model_config = new_model_config

        if hasattr(self.model, "module") and self.model.module is not None:
            self.model.module.model_config = new_model_config

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: Any | None,
    ) -> None:
        del exc_type, exc_val, exc_tb
        self.model.model_config = self.original_config

        if hasattr(self.model, "module") and self.model.module is not None:
            self.model.module.model_config = self.original_config



from colorama import Fore, Style


class DuplicateKeyError(KeyError):
    """Raisable when a mapping (dictionary) key is being duplicated."""


class ExtrasDependencyImportError(Exception):
    """Raisable on ImportErrors due to missing package extras."""

    def __init__(self, extras_type: str, nested_error: Exception):
        message = (
            f"\n\n📦 {nested_error}\n\n"
            + f"{Style.DIM}# have you tried running the following?{Style.RESET_ALL}\n"
            + f"$ {Style.BRIGHT + Fore.GREEN}pip install 'molflux[{extras_type}]'{Style.RESET_ALL}"
        )
        super().__init__(message)



import pandas as pd
import pytest

import datasets
from molflux.modelzoo import load_from_store, save_to_store
from molflux.modelzoo.catalogue import list_models
from molflux.modelzoo.load import load_model
from molflux.modelzoo.models.mapie.mapie_regressor import MapieRegressor
from molflux.modelzoo.protocols import (
    Estimator,
    Model,
    SupportsPredictionInterval,
    SupportsSampling,
    SupportsStandardDeviation,
    SupportsUncertaintyCalibration,
    supports_prediction_interval,
    supports_sampling,
    supports_std,
    supports_uncertainty_calibration,
)

model_name = "mapie_regressor"

_X_FEATURES = ["X_col_1", "X_col_2"]
_Y_FEATURES = ["y_col"]

train_df = pd.DataFrame(
    [
        [1, 2, 10],
        [1, 3, 11],
        [2, 4, 13],
        [2, 4, 13],
        [2, 4, 13],
    ],
    columns=_X_FEATURES + _Y_FEATURES,
)
predict_df = pd.DataFrame(
    [
        [1, 2],
        [1, 3],
        [2, 4],
    ],
    columns=_X_FEATURES,
)
empty_predict_df = pd.DataFrame([], columns=_X_FEATURES)


@pytest.fixture(scope="function")
def fixture_model() -> Model:
    return load_model(model_name, x_features=_X_FEATURES, y_features=_Y_FEATURES)


def test_in_catalogue():
    """That the model is registered in the catalogue."""
    catalogue = list_models()
    all_names = [name for names in catalogue.values() for name in names]
    assert model_name in all_names


def test_default_model_tag_matches_entrypoint_name(fixture_model):
    """That the default model tag includes the catalogue entrypoint name.

    This is not strictly required, but ensures a more intuitive user experience.
    """
    model = fixture_model
    assert model_name in model.tag


def test_is_mapped_to_correct_class(fixture_model):
    """That the model name is mapped to the appropriate class."""
    model = fixture_model
    assert isinstance(model, MapieRegressor)


def test_implements_protocol(fixture_model):
    """That the model implements the public Estimator protocol."""
    model = fixture_model
    assert isinstance(model, Estimator)


def test_implements_uncertainty_calibration_protocol(fixture_model):
    """That the model implements the uncertainty calibration protocol."""
    model = fixture_model
    assert isinstance(model, SupportsUncertaintyCalibration)


def test_implements_prediction_interval_protocol(fixture_model):
    """That the model implements the prediction interval protocol."""
    model = fixture_model
    assert isinstance(model, SupportsPredictionInterval)


def test_implements_standard_deviation_protocol(fixture_model):
    """That the model implements the standard deviation protocol."""
    model = fixture_model
    assert isinstance(model, SupportsStandardDeviation)


def test_implements_sampling_protocol(fixture_model):
    """That the model implements the sampling protocol."""
    model = fixture_model
    assert isinstance(model, SupportsSampling)


def test_tag_for_wrapped_model(fixture_model):
    """That the model has appropriate tag for a wrapped model.
    This is not strictly required, but ensures a more intuitive user experience.
    """
    model = fixture_model
    assert model.tag == "mapie_regressor[linear_regressor]"


def test_raises_on_multitask():
    """That the model cannot be initialised with multiple y features.

    This is not supported yet by the backend model package.
    """
    y_features = ["y1", "y2"]
    with pytest.raises(NotImplementedError, match=r".* single task .*"):
        load_model(model_name, x_features=["x1"], y_features=y_features)


@pytest.mark.parametrize(
    "train_data, predict_data",
    [
        (train_df, predict_df),
        (
            datasets.Dataset.from_pandas(train_df),
            datasets.Dataset.from_pandas(predict_df),
        ),
        (
            datasets.Dataset.from_pandas(train_df),
            datasets.Dataset.from_pandas(empty_predict_df),
        ),
    ],
)
def test_predict_with_prediction_interval_from_estimator(
    fixture_model,
    train_data,
    predict_data,
):
    """That can predict uncertainty as an addon to an existing estimator."""

    existing_estimator = load_model(
        "random_forest_regressor",
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
    )
    existing_estimator.train(train_data)

    # Plug pre-trained model into mapie estimator
    model = load_model(
        name=model_name,
        estimator=existing_estimator,
        x_features=existing_estimator.x_features,
        y_features=existing_estimator.y_features,
        cv="prefit",
    )

    assert supports_uncertainty_calibration(model)
    model.calibrate_uncertainty(data=train_data)

    assert supports_prediction_interval(model)
    predictions, intervals = model.predict_with_prediction_interval(
        predict_data,
        confidence=0.5,
    )

    assert len(predictions) == len(model.y_features)
    for task_predictions in predictions.values():
        assert len(task_predictions) == len(predict_data)

    assert len(intervals) == len(model.y_features)
    for prediction_interval in intervals.values():
        assert len(prediction_interval) == len(predict_data)


@pytest.mark.parametrize(
    "train_data, predict_data",
    [
        (train_df, predict_df),
        (
            datasets.Dataset.from_pandas(train_df),
            datasets.Dataset.from_pandas(predict_df),
        ),
        (
            datasets.Dataset.from_pandas(train_df),
            datasets.Dataset.from_pandas(empty_predict_df),
        ),
    ],
)
def test_train_predict_with_standard_deviation(fixture_model, train_data, predict_data):
    """Test that a model can run the train and predict_with_std functions"""
    model = fixture_model
    model.train(train_data)

    assert supports_std(model)

    predictions, stds = model.predict_with_std(predict_data)

    assert len(predictions) == len(model.y_features)
    for task_predictions in predictions.values():
        assert len(task_predictions) == len(predict_data)

    assert len(stds) == len(model.y_features)
    for prediction_std in stds.values():
        assert len(prediction_std) == len(predict_data)


@pytest.mark.parametrize(
    "train_data, predict_data",
    [
        (train_df, predict_df),
        (
            datasets.Dataset.from_pandas(train_df),
            datasets.Dataset.from_pandas(predict_df),
        ),
        (
            datasets.Dataset.from_pandas(train_df),
            datasets.Dataset.from_pandas(empty_predict_df),
        ),
    ],
)
def test_sample(fixture_model, train_data, predict_data):
    """Test that a model can run the train and sample functions"""
    model = fixture_model
    model.train(train_data)

    assert supports_sampling(model)

    n_samples = 10
    prediction_samples = model.sample(predict_data, n_samples=n_samples)
    assert prediction_samples
    assert len(prediction_samples) == len(model.y_features)
    for samples in prediction_samples.values():
        assert len(samples) == len(predict_data)
        for item in samples:
            assert len(item) == n_samples


def test_saving_mapie_regressor_model(tmp_path):
    """That can save a mapie regressor model"""

    existing_estimator = load_model(
        "random_forest_regressor",
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
    )
    existing_estimator.train(train_df)

    # Plug pre-trained model into mapie estimator
    model = load_model(
        name=model_name,
        estimator=existing_estimator,
        x_features=existing_estimator.x_features,
        y_features=existing_estimator.y_features,
        cv="prefit",
    )

    model.calibrate_uncertainty(data=train_df)  # type: ignore[attr-defined]
    save_to_store(tmp_path, model)

    loaded_model = load_from_store(tmp_path)

    assert loaded_model.predict_with_prediction_interval(predict_df, confidence=0.5)  # type: ignore[attr-defined]


def test_loading_pretrained_raises_on_training_attempt(tmp_path):
    """That an error is raised if attempting to retrain a pre-trained model that
    loaded from a persisted artefact.

    This test should be removed if we end up finding a way to save models in
    a way that persists a link to the input estimator wrapped by the MAPIE
    model.
    """

    existing_estimator = load_model(
        "random_forest_regressor",
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
    )
    existing_estimator.train(train_df)

    # Plug pre-trained model into mapie estimator
    model = load_model(
        name=model_name,
        estimator=existing_estimator,
        x_features=existing_estimator.x_features,
        y_features=existing_estimator.y_features,
        cv="prefit",
    )

    model.calibrate_uncertainty(data=train_df)  # type: ignore[attr-defined]
    save_to_store(tmp_path, model)

    with pytest.warns(
        UserWarning,
        match="Model loaded with an unlinked input estimator",
    ):
        loaded_model = load_from_store(tmp_path)

    with pytest.raises(ValueError, match="The input estimator has been unlinked"):
        loaded_model.calibrate_uncertainty(train_df)  # type: ignore[attr-defined]


def test_predict_single_instance(fixture_model):
    """Test that predicting on a single data point behaves as expected"""
    train_data = datasets.Dataset.from_pandas(train_df)
    predict_data = datasets.Dataset.from_pandas(predict_df.iloc[:1])

    model = fixture_model
    model.train(train_data)

    predictions = model.predict(predict_data)
    for array in predictions.values():
        assert isinstance(array, list)

    predictions, prediction_intervals = model.predict_with_prediction_interval(
        predict_data,
        confidence=0.5,
    )
    for array in predictions.values():
        assert isinstance(array, list)
    for array in prediction_intervals.values():
        assert isinstance(array, list)
        assert isinstance(array[0], tuple)


@pytest.mark.parametrize(
    ["estimator", "should_raise_errors"],
    [
        [
            {
                "name": "random_forest_regressor",
                "config": {"n_estimators": 10, "x_features": ["dummy_col"]},
            },
            False,
        ],
        [{"name": "gradient_boosting_regressor"}, False],
        [{"bad_schema_key": {"bad_kwarg": 42}}, True],
        [{"name": "bad_name_for_a_model"}, True],
    ],
)
def test_model_with_config_estimator(estimator, should_raise_errors):
    """That models with configs as estimators correctly run"""

    model: Model = load_model(
        model_name,
        estimator=estimator,
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
    )

    train_data = datasets.Dataset.from_pandas(train_df)
    predict_data = datasets.Dataset.from_pandas(predict_df)

    if should_raise_errors:
        with pytest.raises(ValueError):
            model.train(train_data)
        return
    else:
        model.train(train_data)

    assert isinstance(model, SupportsPredictionInterval)

    predictions, prediction_intervals = model.predict_with_prediction_interval(
        predict_data,
        confidence=0.5,
    )
    for array in predictions.values():
        assert isinstance(array, list)
    for array in prediction_intervals.values():
        assert isinstance(array, list)
        assert isinstance(array[0], tuple)



import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.representations.core.generic.sum import Sum

representation_name = "sum"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, Sum)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_default_representation_tag_matches_entrypoint_name(fixture_representation):
    """That the default representation tag matches the entrypoint name.

    This is not strictly required, but ensures a more intuitive user experience.
    """
    representation = fixture_representation
    assert representation.tag == representation_name


@pytest.mark.parametrize(
    "columns,expected_output_column",
    [
        (
            [
                [0, 1, 2],
                [3, 4, 5],
                [6, 7, 8],
                [0, 1, 2],
            ],
            [
                9,
                13,
                17,
            ],
        ),
        (
            [
                [0],
                [1],
                [2],
            ],
            [
                3,
            ],
        ),
        (
            [
                [],
                [],
                [],
            ],
            [],
        ),
    ],
)
def test_sum_features(fixture_representation, columns, expected_output_column):
    """That works as expected on input columns"""
    representation = fixture_representation

    results = representation.featurise(*columns)

    # the only key in the output should be the representation tag
    assert [fixture_representation.tag] == list(results.keys())

    # check that the expected sums match
    assert expected_output_column == results[fixture_representation.tag]



"""Recall score."""

import logging
from typing import Any

import evaluate
from sklearn.metrics import recall_score

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Compute the recall score.
The recall is the ratio ``tp / (tp + fn)`` where ``tp`` is the number of
true positives and ``fn`` the number of false negatives. The recall is
intuitively the ability of the classifier to find all the positive samples.
The best value is 1 and the worst value is 0.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated targets as returned by a classifier.
    references: Ground truth (correct) target values.
    average (optional): {'micro', 'macro', 'samples','weighted', 'binary'} or
        None, default='binary'. This parameter is required for
        multiclass/multilabel targets. If ``None``, the scores for each class
        are returned. Otherwise, this determines the type of averaging performed
        on the data:

        ``'binary'``:
            Only report results for the class specified by ``pos_label``.
            This is applicable only if targets (``y_{true,pred}``) are binary.
        ``'micro'``:
            Calculate metrics globally by counting the total true positives,
            false negatives and false positives.
        ``'macro'``:
            Calculate metrics for each label, and find their unweighted
            mean.  This does not take label imbalance into account.
        ``'weighted'``:
            Calculate metrics for each label, and find their average weighted
            by support (the number of true instances for each label). This
            alters 'macro' to account for label imbalance; it can result in an
            F-score that is not between precision and recall.
        ``'samples'``:
            Calculate metrics for each instance, and find their average (only
            meaningful for multilabel classification where this differs from
            :func:`accuracy_score`).
    labels (optional): The set of labels to include when
        ``average != 'binary'``, and their order if ``average is None``.
        Labels present in the data can be excluded, for example to calculate a
        multiclass average ignoring a majority negative class, while labels not
        present in the data will result in 0 components in a macro average.
        For multilabel targets, labels are column indices. By default, all
        labels in ``y_true`` and ``y_pred`` are used in sorted order.
    pos_label (optional): The class to report if ``average='binary'`` and the
        data is binary. If the data are multiclass or multilabel, this will be
        ignored; setting ``labels=[pos_label]`` and ``average != 'binary'``
        will report scores for that label only. Defaults to 1.
    sample_weight (optional): Weighting of each sample.
    zero_division (optional): "warn", 0 or 1. Sets the value to return when
        there is a zero division, i.e. when all predictions and labels are
        negative. If set to "warn", this acts as 0, but warnings are also
        raised. Defaults to "warn".

Returns:
    recall: Recall of the positive class in binary classification or
    weighted average of the precision of each class for the multiclass task.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("recall")
    >>> references = [0, 1, 2, 0, 1, 2]
    >>> predictions = [0, 2, 1, 0, 0, 1]
    >>> metric.compute(references=references, predictions=predictions, average="macro")
    {'recall': 0.33...}
    >>> metric.compute(references=references, predictions=predictions, average="micro")
    {'recall': 0.33...}
    >>> metric.compute(references=references, predictions=predictions, average="weighted")
    {'recall': 0.33...}
    >>> metric.compute(references=references, predictions=predictions, average=None)
    {'recall': [1.0, 0.0, 0.0]}
    >>> references = [0, 0, 0, 0, 0, 0]
    >>> metric.compute(references=references, predictions=predictions, average=None, zero_division=0.0)
    {'recall': [0.5, 0.0, 0.0]}
    >>> metric.compute(references=references, predictions=predictions, average=None, zero_division=1.0)
    {'recall': [0.5, 1.0, 1.0]}

Notes:
    When ``true positive + false negative == 0``, recall returns 0 and raises
    ``UndefinedMetricWarning``. This behavior can be modified with
    ``zero_division``.
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class Recall(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("int32")),
                    "references": datasets.Sequence(datasets.Value("int32")),
                }
                if self.config_name == "multilabel"
                else {
                    "predictions": datasets.Value("int32"),
                    "references": datasets.Value("int32"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.recall_score.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: list[float] | None = None,
        labels: list[str] | None = None,
        pos_label: int = 1,
        average: str | None = "binary",
        zero_division: int | str = "warn",
        **kwargs: Any,
    ) -> MetricResult:
        score = recall_score(
            y_true=references,
            y_pred=predictions,
            labels=labels,
            pos_label=pos_label,
            average=average,
            sample_weight=sample_weight,
            zero_division=zero_division,
        )
        return {self.tag: score.tolist()}



from typing import Literal

from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import (
    SKLearnClassificationMixin,
    SKLearnModelBase,
)

try:
    from sklearn.dummy import DummyClassifier as SKDummyClassifier
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
This is an sklearn dummy classifier model.

This classifier serves as a simple baseline to compare against other more complex classifiers. The specific behavior of
the baseline is selected with the strategy parameter. All strategies make predictions that ignore the input x feature
values provided at training and prediction time. The predictions, however, typically depend on values observed in the
y features provided during training.

Note that the “stratified” and “uniform” strategies lead to non-deterministic predictions that can be rendered
deterministic by setting the random_state parameter if needed. The other strategies are naturally deterministic and,
once trained, always return the same constant prediction for any value of input x features.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
strategy : {"most_frequent", "prior", "stratified", "uniform", "constant"}, default="prior"
    Strategy to use to generate predictions.

    * "most_frequent": the `predict` method always returns the most
      frequent class label in the observed y features passed during training.
      The `predict_proba` method returns the matching one-hot encoded
      vector.
    * "prior": the `predict` method always returns the most frequent
      class label in the observed y features passed during training (like
      "most_frequent"). ``predict_proba`` always returns the empirical
      class distribution of `y` also known as the empirical class prior
      distribution.
    * "stratified": the `predict_proba` method randomly samples one-hot
      vectors from a multinomial distribution parametrized by the empirical
      class prior probabilities.
      The `predict` method returns the class label which got probability
      one in the one-hot vector of `predict_proba`.
      Each sampled row of both methods is therefore independent and
      identically distributed.
    * "uniform": generates predictions uniformly at random from the list
      of unique classes observed in `y`, i.e. each class has equal
      probability.
    * "constant": always predicts a constant label that is provided by
      the user. This is useful for metrics that evaluate a non-majority
      class.

random_state : int or None, default=None
    Controls the randomness to generate the predictions when
    ``strategy='stratified'`` or ``strategy='uniform'``.
    Pass an int for reproducible output across multiple function calls.

constant : int or str or array-like of shape (n_outputs,), default=None
    The explicit constant as predicted by the "constant" strategy. This
    parameter is useful only for the "constant" strategy.
"""

PredictionsStrategy = Literal[
    "most_frequent",
    "prior",
    "stratified",
    "uniform",
    "constant",
]


class Config:
    extra = "forbid"
    arbitrary_types_allowed = True
    smart_union = True


@dataclasses.dataclass(config=Config)
class DummyClassifierConfig(ModelConfig):
    strategy: PredictionsStrategy = "prior"
    random_state: int | None = None
    constant: int | str | list[int | str] | None = None


class DummyClassifier(
    SKLearnClassificationMixin,
    SKLearnModelBase[DummyClassifierConfig],
):
    @property
    def _config_builder(self) -> type[DummyClassifierConfig]:
        return DummyClassifierConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKDummyClassifier:
        config = self.model_config
        return SKDummyClassifier(
            strategy=config.strategy,
            random_state=config.random_state,
            constant=config.constant,
        )



import logging
from typing import Any

try:
    from rdkit.Chem.rdmolops import PatternFingerprint
    from rdkit.DataStructs.cDataStructs import ExplicitBitVect

    from molflux.features.representations.rdkit._utils import rdkit_mol_from_smiles
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

logger = logging.getLogger(__name__)

_DESCRIPTION = """
A fingerprint using SMARTS patterns.

Topological fingerprint for a molecule are generated using a series of
pre-defined structural patterns.

These fingerprints were designed to be used in substructure screening. These
are, as far as we know, unique to the RDKit. The algorithm identifies features
in the molecule by doing substructure searches using a small number
(12 in the 2019.03 release of the RDKit) of very generic SMARTS patterns -
like `[*]~[*]~[*](~[*])~[*]` or `[R]~1[R]~[R]~[R]~1`, and then hashing each
occurrence of a pattern based on the atom and bond types involved. The fact
that particular pattern matched the molecule at all is also stored by hashing t
he pattern ID and size. If a particular feature contains either a query atom or
a query bond (e.g. something generated from SMARTS), the only information that
is hashed is the fact that the generic pattern matched.

For the 2019.03 release, the atom types use just the atomic number of the atom
and the bond types use the bond type, or AROMATIC for aromatic bonds).

NOTE: Because it plays an important role in substructure screenout, the
internals of this fingerprint (the generic patterns used and/or the details of
the hashing algorithm) may change from one release to the next.
"""


class Pattern(RepresentationBase):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        logging.warning(
            "Pattern fingerprints are experimental and may change with rdkit updates.",
        )

    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        fp_size: int = 2048,
        atom_counts: list[int] | None = None,
        set_only_bits: ExplicitBitVect | None = None,
        tautomer_fingerprints: bool = False,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Generates topological fingerprints for each input molecule using a
        series of pre-defined structural patterns.

        Args:
            samples: The molecules to be fingerprinted.
            fp_size: The size of the fingerprint. Defaults to `2048`
            atom_counts: If provided, this will be used to provide the count
                of the number of paths that set bits each atom is involved in.
                The vector should have at least as many entries as the molecule
                has atoms and is not zeroed out here.
            set_only_bits: If provided, only bits that are set in this bit
                vector will be set in the result. This is essentially the same
                as doing: `(*res) &= (*setOnlyBits)`; but also has an impact on
                the `atom_counts` (if being used). Defaults to `None`.
            tautomer_fingerprints: Defaults to `False`.

        Returns:
            Pattern fingerprints, as lists of bits.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('pattern')
            >>> samples = ['c1ccccc1']
            >>> representation.featurise(samples, fp_size=16)  # doctest: +SKIP
            {'pattern': [[0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        pattern_fp_list: list[list] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smile = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smile)
                rd_fp = PatternFingerprint(
                    mol,
                    fpSize=fp_size,
                    atomCounts=atom_counts or [],
                    setOnlyBits=set_only_bits,
                    tautomerFingerprints=tautomer_fingerprints,
                )

                if rd_fp.GetNumOnBits() == 0:
                    pattern_fp_list.append([0] * rd_fp.GetNumBits())
                else:
                    pattern_fp_list.append(rd_fp.ToList())

        return {self.tag: pattern_fp_list}



from collections.abc import Iterator
from typing import Any, Protocol, runtime_checkable

from molflux.features.errors import DuplicateKeyError
from molflux.features.typing import ArrayLike, RepresentationResult


@runtime_checkable
class Representation(Protocol):
    """The public protocol for molflux.features.Representations."""

    def __init__(self, **kwargs: Any):
        """Initialises the representation."""

    @property
    def metadata(self) -> dict[str, Any]:
        """Representation metadata."""

    @property
    def name(self) -> str:
        """The canonical name of the representation."""

    @property
    def tag(self) -> str:
        """The arbitrary tag name of the representation."""

    @property
    def state(self) -> dict[str, Any]:
        """The internal state config."""

    def featurise(self, *columns: ArrayLike, **kwargs: Any) -> RepresentationResult:
        """Featurises the input samples."""

    def reset_state(self) -> None:
        """Resets the state."""

    def update_state(self, **kwargs: Any) -> None:
        """Updates the internal state config."""


class Representations:
    """A collection of Representations."""

    def __init__(self) -> None:
        self._stack: dict[str, Representation] = {}

    def __contains__(self, item: str) -> bool:
        return item in self._stack

    def __getitem__(self, key: str) -> Representation:
        representation = self._stack.get(key)
        if representation is None:
            raise KeyError(
                f"Representation {key!r} has not been loaded. Available representations: {list(self._stack.keys())!r}",
            )
        return representation

    def __iter__(self) -> Iterator[Representation]:
        """Iterates through Representations as if a list."""
        return iter(self._stack.values())

    def __len__(self) -> int:
        return len(self._stack)

    def __repr__(self) -> str:
        tags = sorted(representation.tag for representation in self._stack.values())
        return f"Representations({tags!r})"

    def __setitem__(self, key: str, value: Representation) -> None:
        if key in self._stack:
            raise DuplicateKeyError(
                f"Representation with key {key!r} has already been added. You can add a unique custom tag to one of the representations to add it to the collection.",
            )
        self._stack[key] = value

    def add_representation(self, representation: Representation) -> None:
        self[representation.tag] = representation

    def featurise(self, *columns: ArrayLike, **kwargs: Any) -> RepresentationResult:
        results = (
            representation.featurise(*columns)
            for representation in self._stack.values()
        )
        merged_results = {k: v for r in results for k, v in r.items()}
        return merged_results



if __name__ == "__main__":
    import h5py

    import datasets

    dl_manager = datasets.DownloadManager()

    h5_file_path = dl_manager.download_and_extract(
        "https://zenodo.org/records/8222043/files/SPICE-1.1.4.hdf5?download=1",
    )

    h5_file = h5py.File(h5_file_path, "r")

    new_h5_file = h5py.File("data.h5", "w")

    for key in list(h5_file.keys())[:5]:
        new_h5_file.create_group(key)
        new_h5_file[key].create_dataset(
            "atomic_numbers",
            data=h5_file[key]["atomic_numbers"],
        )
        new_h5_file[key].create_dataset("smiles", data=h5_file[key]["smiles"])
        new_h5_file[key].create_dataset("subset", data=h5_file[key]["subset"])

        new_h5_file[key].create_dataset(
            "conformations",
            data=h5_file[key]["conformations"][:2],
        )
        new_h5_file[key].create_dataset(
            "dft_total_energy",
            data=h5_file[key]["dft_total_energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "dft_total_gradient",
            data=h5_file[key]["dft_total_gradient"][:2],
        )
        new_h5_file[key].create_dataset(
            "formation_energy",
            data=h5_file[key]["formation_energy"][:2],
        )
        new_h5_file[key].create_dataset(
            "mbis_dipoles",
            data=h5_file[key]["mbis_dipoles"][:2],
        )
        new_h5_file[key].create_dataset(
            "mbis_quadrupoles",
            data=h5_file[key]["mbis_quadrupoles"][:2],
        )
        new_h5_file[key].create_dataset(
            "mbis_octupoles",
            data=h5_file[key]["mbis_octupoles"][:2],
        )
        new_h5_file[key].create_dataset(
            "scf_dipole",
            data=h5_file[key]["scf_dipole"][:2],
        )
        new_h5_file[key].create_dataset(
            "scf_quadrupole",
            data=h5_file[key]["scf_quadrupole"][:2],
        )
        new_h5_file[key].create_dataset(
            "mayer_indices",
            data=h5_file[key]["mayer_indices"][:2],
        )
        new_h5_file[key].create_dataset(
            "wiberg_lowdin_indices",
            data=h5_file[key]["wiberg_lowdin_indices"][:2],
        )
    new_h5_file.close()



from typing import Any

try:
    from openeye.oechem import OEFloatArray
    from openeye.oemolprop import OEGetXLogPResult
    from openeye.oequacpac import OERemoveFormalCharge
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import to_digit, to_oemol
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
The XLogP algorithm ([Wang-1997-2]) is provided because its atom-type
contribution allows calculation of the XLogP contribution of any fragment in a
molecule and allows minimal corrections in a simple additive form to calculate
the LogP of any molecule made from combinations of fragments.

Further, although the method contains many many free parameters, its simple
linear form allows for ready interpretation of the model and most of the
parameters in the model make rational sense.

Unfortunately, the original algorithm is difficult to implement as published.
First, the internal-hydrogen bond term was calculated using a single
3D conformation. It was found that this was both arbitrary and unnecessary.
This arbitrary 3D calculation has been replaced with a 2D approach to recognize
common internal-hydrogen bonds. In tests, this 2D method worked comparably to
the published 3D algorithm. Next, the training set had a few subtle atom-type
inconsistencies.
"""


class XLogP(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        digitise: bool = False,
        start: int = -3,
        stop: int = 7,
        num: int = 10,
        atom_xlogps: OEFloatArray | None = None,
        remove_formal_charge: bool = True,
        **kwargs: Any,
    ) -> dict[str, list[float]]:
        """Calculates the XLogP value for each input molecule.

        .. hint::

            XLogP should be calculated on the neutral form of the molecule.
            Therefore, calling it is highly recommended to make sure to set
            `remove_formal_charge` to `True` when using this representation.

        Args:
            samples: The sample molecules to featurise.
            digitise: Whether to digitise the results. Defaults to `False`.
            start: If `digitise=True`, the start of the binning range. Defaults
                to -3.
            stop: If `digitise=True`, the end of the binning range. Defaults
                to 7.
            num: If `digitise=True`, the number of bins. Defaults to 10.
            atom_xlogps: Can be used to retrieve the contribution of each atom
                to the total XLogP.
            remove_formal_charge: Whether to attempt to remove all formal
                charges from mol in a manner that is consistent with adding and
                removing implicit or explicit protons. This method will not
                create radicals and will not attach more protons to an atom
                than is acceptable in that atom`s standard valence form.
                Please note that the formal charge of quaternary amines is not
                removed with this flag. Defaults to `True`.

        Returns:
            The computed XLogP values on each input sample.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('x_log_p')
            >>> samples = ['C1=CC=C(C=C1)S(=O)O', 'C']
            >>> representation.featurise(samples=samples)
            {'x_log_p': [-0.31300002336502075, 0.27699998021125793]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        results = []
        for sample in samples:
            with featurisation_error_harness(sample):
                mol = to_oemol(sample)
                # XLogP should be calculated on the neutral form of the molecule
                if remove_formal_charge:
                    OERemoveFormalCharge(mol)

                result = OEGetXLogPResult(mol, atom_xlogps).GetValue()

                if digitise:
                    result = to_digit(result, start=start, stop=stop, num=num).tolist()

                results.append(result)

        return {self.tag: results}



from os import PathLike as OSPathLike
from typing import (
    Any,
    Protocol,
    Union,
    runtime_checkable,
)

import datasets


@runtime_checkable
class SupportsDataframeInterchangeProtocol(Protocol):
    """Protocol representing any object conforming to the Dataframe Interchange Protocol.

    Several DataFrame libraries now support this interchange protocol (pandas, modin, vaex, cudf, pyarrow).

    References:
        https://data-apis.org/dataframe-protocol/latest/purpose_and_scope.html
        https://data-apis.org/dataframe-protocol/latest/API.html
    """

    def __dataframe__(
        self,
        nan_as_null: bool = False,
        allow_copy: bool = True,
    ) -> Any: ...


DataFrameLike = Union[SupportsDataframeInterchangeProtocol, datasets.Dataset]
Classes = dict[str, list[Any]]
Features = list[str]
PathLike = Union[str, OSPathLike]
PredictionResult = dict[str, list[Any]]



from typing import Any

from molflux.features.utils import assert_n_positional_args

try:
    from rdkit.Avalon.pyAvalonTools import GetAvalonCountFP
    from rdkit.Chem import DataStructs, rdReducedGraphs
    from rdkit.Chem.rdMolDescriptors import GetHashedMorganFingerprint
    from rdkit.ML.Descriptors.MoleculeDescriptors import MolecularDescriptorCalculator

    from molflux.features.representations.rdkit._utils import rdkit_mol_from_smiles
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

import numpy as np

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import featurisation_error_harness

_DESCRIPTION = """
MapLight features, as detailed in the ADMET property prediction through combinations of molecular fingerprints.

Source code adapted to molflux: https://github.com/maplightrx/MapLight-TDC/blob/main/maplight.py
"""


class MapLight(RepresentationBase):
    OUTPUT_SIZE: int = 2563

    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Generates the MapLight features for each input molecule.
        This includes 1024 Morgan bits, 1024 Avalon bits, 315 Reduced Graph bits and 200 descriptors, for a total
        of 2563 features.

        Args:
            samples: The molecules to be featurised.

        Returns:
            MapLight features, as lists of floats.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('map_light')
            >>> samples = ['c1ccccc1']
            >>> representation.featurise(samples)  # doctest:+ELLIPSIS
            {'map_light': [[0.0, 0.0, 0.0, 0.0, ...
        """

        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]

        morgan_fingerprint_list = []
        avalon_fingerprint_list = []
        erg_fingerprint_list = []
        rdkit_features_list = []

        calculator = MolecularDescriptorCalculator(get_chosen_descriptors())

        for sample in samples:
            with featurisation_error_harness(sample):
                smile = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smile)

                # compute the Morgan Fingerprint
                morgan_fp = GetHashedMorganFingerprint(mol, nBits=1024, radius=2)
                morgan_fp = count_to_array(morgan_fp)
                morgan_fingerprint_list.append(morgan_fp)

                # compute the Avalon Fingerprint
                avalon_fp = GetAvalonCountFP(mol, nBits=1024)
                avalon_fp = count_to_array(avalon_fp)
                avalon_fingerprint_list.append(avalon_fp)

                # compute the Reduced Graph Fingerprint
                erg_fp = rdReducedGraphs.GetErGFingerprint(mol)
                erg_fingerprint_list.append(erg_fp)

                # compute the RDKit descriptors
                rdkit_features = calculator.CalcDescriptors(mol)
                rdkit_features = np.array(rdkit_features)
                rdkit_features_list.append(rdkit_features)

        # stack the features vertically in 4 2-dimensional numpy arrays
        fingerprints = [
            np.vstack(morgan_fingerprint_list),
            np.vstack(avalon_fingerprint_list),
            np.vstack(erg_fingerprint_list),
            np.vstack(rdkit_features_list),
        ]

        # concatenate the features along the column axis
        result = np.concatenate(fingerprints, axis=1)

        return {self.tag: result.tolist()}


def count_to_array(fingerprint: Any) -> np.ndarray:
    array = np.zeros((0,), dtype=np.int8)

    DataStructs.ConvertToNumpyArray(fingerprint, array)

    return array


# from https://www.blopig.com/blog/2022/06/how-to-turn-a-molecule-into-a-vector-of-physicochemical-descriptors-using-rdkit/
def get_chosen_descriptors() -> list[str]:
    """Simple function for returning a list of handcrafted rdkit features"""
    chosen_descriptors = [
        "BalabanJ",
        "BertzCT",
        "Chi0",
        "Chi0n",
        "Chi0v",
        "Chi1",
        "Chi1n",
        "Chi1v",
        "Chi2n",
        "Chi2v",
        "Chi3n",
        "Chi3v",
        "Chi4n",
        "Chi4v",
        "EState_VSA1",
        "EState_VSA10",
        "EState_VSA11",
        "EState_VSA2",
        "EState_VSA3",
        "EState_VSA4",
        "EState_VSA5",
        "EState_VSA6",
        "EState_VSA7",
        "EState_VSA8",
        "EState_VSA9",
        "ExactMolWt",
        "FpDensityMorgan1",
        "FpDensityMorgan2",
        "FpDensityMorgan3",
        "FractionCSP3",
        "HallKierAlpha",
        "HeavyAtomCount",
        "HeavyAtomMolWt",
        "Ipc",
        "Kappa1",
        "Kappa2",
        "Kappa3",
        "LabuteASA",
        "MaxAbsEStateIndex",
        "MaxAbsPartialCharge",
        "MaxEStateIndex",
        "MaxPartialCharge",
        "MinAbsEStateIndex",
        "MinAbsPartialCharge",
        "MinEStateIndex",
        "MinPartialCharge",
        "MolLogP",
        "MolMR",
        "MolWt",
        "NHOHCount",
        "NOCount",
        "NumAliphaticCarbocycles",
        "NumAliphaticHeterocycles",
        "NumAliphaticRings",
        "NumAromaticCarbocycles",
        "NumAromaticHeterocycles",
        "NumAromaticRings",
        "NumHAcceptors",
        "NumHDonors",
        "NumHeteroatoms",
        "NumRadicalElectrons",
        "NumRotatableBonds",
        "NumSaturatedCarbocycles",
        "NumSaturatedHeterocycles",
        "NumSaturatedRings",
        "NumValenceElectrons",
        "PEOE_VSA1",
        "PEOE_VSA10",
        "PEOE_VSA11",
        "PEOE_VSA12",
        "PEOE_VSA13",
        "PEOE_VSA14",
        "PEOE_VSA2",
        "PEOE_VSA3",
        "PEOE_VSA4",
        "PEOE_VSA5",
        "PEOE_VSA6",
        "PEOE_VSA7",
        "PEOE_VSA8",
        "PEOE_VSA9",
        "RingCount",
        "SMR_VSA1",
        "SMR_VSA10",
        "SMR_VSA2",
        "SMR_VSA3",
        "SMR_VSA4",
        "SMR_VSA5",
        "SMR_VSA6",
        "SMR_VSA7",
        "SMR_VSA8",
        "SMR_VSA9",
        "SlogP_VSA1",
        "SlogP_VSA10",
        "SlogP_VSA11",
        "SlogP_VSA12",
        "SlogP_VSA2",
        "SlogP_VSA3",
        "SlogP_VSA4",
        "SlogP_VSA5",
        "SlogP_VSA6",
        "SlogP_VSA7",
        "SlogP_VSA8",
        "SlogP_VSA9",
        "TPSA",
        "VSA_EState1",
        "VSA_EState10",
        "VSA_EState2",
        "VSA_EState3",
        "VSA_EState4",
        "VSA_EState5",
        "VSA_EState6",
        "VSA_EState7",
        "VSA_EState8",
        "VSA_EState9",
        "fr_Al_COO",
        "fr_Al_OH",
        "fr_Al_OH_noTert",
        "fr_ArN",
        "fr_Ar_COO",
        "fr_Ar_N",
        "fr_Ar_NH",
        "fr_Ar_OH",
        "fr_COO",
        "fr_COO2",
        "fr_C_O",
        "fr_C_O_noCOO",
        "fr_C_S",
        "fr_HOCCN",
        "fr_Imine",
        "fr_NH0",
        "fr_NH1",
        "fr_NH2",
        "fr_N_O",
        "fr_Ndealkylation1",
        "fr_Ndealkylation2",
        "fr_Nhpyrrole",
        "fr_SH",
        "fr_aldehyde",
        "fr_alkyl_carbamate",
        "fr_alkyl_halide",
        "fr_allylic_oxid",
        "fr_amide",
        "fr_amidine",
        "fr_aniline",
        "fr_aryl_methyl",
        "fr_azide",
        "fr_azo",
        "fr_barbitur",
        "fr_benzene",
        "fr_benzodiazepine",
        "fr_bicyclic",
        "fr_diazo",
        "fr_dihydropyridine",
        "fr_epoxide",
        "fr_ester",
        "fr_ether",
        "fr_furan",
        "fr_guanido",
        "fr_halogen",
        "fr_hdrzine",
        "fr_hdrzone",
        "fr_imidazole",
        "fr_imide",
        "fr_isocyan",
        "fr_isothiocyan",
        "fr_ketone",
        "fr_ketone_Topliss",
        "fr_lactam",
        "fr_lactone",
        "fr_methoxy",
        "fr_morpholine",
        "fr_nitrile",
        "fr_nitro",
        "fr_nitro_arom",
        "fr_nitro_arom_nonortho",
        "fr_nitroso",
        "fr_oxazole",
        "fr_oxime",
        "fr_para_hydroxylation",
        "fr_phenol",
        "fr_phenol_noOrthoHbond",
        "fr_phos_acid",
        "fr_phos_ester",
        "fr_piperdine",
        "fr_piperzine",
        "fr_priamide",
        "fr_prisulfonamd",
        "fr_pyridine",
        "fr_quatN",
        "fr_sulfide",
        "fr_sulfonamd",
        "fr_sulfone",
        "fr_term_acetylene",
        "fr_tetrazole",
        "fr_thiazole",
        "fr_thiocyan",
        "fr_thiophene",
        "fr_unbrch_alkane",
        "fr_urea",
        "qed",
    ]

    return chosen_descriptors



from typing import Literal, Optional

from numpy.random import RandomState
from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import (
    SKLearnClassificationMixin,
    SKLearnModelBase,
)

try:
    from sklearn.linear_model import LogisticRegression as SKLogisticRegression
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
This is an sklearn logistic regression model.

Logistic regression, despite its name, is a linear model for classification rather
than regression. Logistic regression is also known in the literature as logit
regression, maximum-entropy classification (MaxEnt) or the log-linear classifier.
In this model, the probabilities describing the possible outcomes of a single
trial are modeled using a logistic function.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
penalty : {'l1', 'l2', 'elasticnet', 'none'}, default='l2'
    Specify the norm of the penalty:
    - `'none'`: no penalty is added;
    - `'l2'`: add a L2 penalty term and it is the default choice;
    - `'l1'`: add a L1 penalty term;
    - `'elasticnet'`: both L1 and L2 penalty terms are added.
    .. warning::
       Some penalties may not work with some solvers. See the parameter
       `solver` below, to know the compatibility between the penalty and
       solver.
dual : bool, default=False
    Dual or primal formulation. Dual formulation is only implemented for
    l2 penalty with liblinear solver. Prefer dual=False when
    n_samples > n_features.
tol : float, default=1e-4
    Tolerance for stopping criteria.
C : float, default=1.0
    Inverse of regularization strength; must be a positive float.
    Like in support vector machines, smaller values specify stronger
    regularization.
fit_intercept : bool, default=True
    Specifies if a constant (a.k.a. bias or intercept) should be
    added to the decision function.
intercept_scaling : float, default=1
    Useful only when the solver 'liblinear' is used
    and self.fit_intercept is set to True. In this case, x becomes
    [x, self.intercept_scaling],
    i.e. a "synthetic" feature with constant value equal to
    intercept_scaling is appended to the instance vector.
    The intercept becomes ``intercept_scaling * synthetic_feature_weight``.
    Note! the synthetic feature weight is subject to l1/l2 regularization
    as all other features.
    To lessen the effect of regularization on synthetic feature weight
    (and therefore on the intercept) intercept_scaling has to be increased.
class_weight : dict or 'balanced', default=None
    Weights associated with classes in the form ``{class_label: weight}``.
    If not given, all classes are supposed to have weight one.
    The "balanced" mode uses the values of y to automatically adjust
    weights inversely proportional to class frequencies in the input data
    as ``n_samples / (n_classes * np.bincount(y))``.
    Note that these weights will be multiplied with sample_weight (passed
    through the fit method) if sample_weight is specified.
random_state : int, RandomState instance, default=None
    Used when ``solver`` == 'sag', 'saga' or 'liblinear' to shuffle the
    data.
solver : {'newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga'}, \
        default='lbfgs'
    Algorithm to use in the optimization problem. Default is 'lbfgs'.
    To choose a solver, you might want to consider the following aspects:
        - For small datasets, 'liblinear' is a good choice, whereas 'sag'
          and 'saga' are faster for large ones;
        - For multiclass problems, only 'newton-cg', 'sag', 'saga' and
          'lbfgs' handle multinomial loss;
        - 'liblinear' is limited to one-versus-rest schemes.
    .. warning::
       The choice of the algorithm depends on the penalty chosen:
       Supported penalties by solver:
       - 'newton-cg'   -   ['l2', 'none']
       - 'lbfgs'       -   ['l2', 'none']
       - 'liblinear'   -   ['l1', 'l2']
       - 'sag'         -   ['l2', 'none']
       - 'saga'        -   ['elasticnet', 'l1', 'l2', 'none']
    .. note::
       'sag' and 'saga' fast convergence is only guaranteed on
       features with approximately the same scale. You can
       preprocess the data with a scaler from :mod:`sklearn.preprocessing`.
    .. seealso::
       Refer to the User Guide for more information regarding
       :class:`LogisticRegression` and more specifically the
       `Table <https://scikit-learn.org/dev/modules/linear_model.html#logistic-regression>`_
       summarazing solver/penalty supports.
max_iter : int, default=100
    Maximum number of iterations taken for the solvers to converge.
verbose : int, default=0
    For the liblinear and lbfgs solvers set verbose to any positive
    number for verbosity.
warm_start : bool, default=False
    When set to True, reuse the solution of the previous call to fit as
    initialization, otherwise, just erase the previous solution.
    Useless for liblinear solver.
l1_ratio : float, default=None
    The Elastic-Net mixing parameter, with ``0 <= l1_ratio <= 1``. Only
    used if ``penalty='elasticnet'``. Setting ``l1_ratio=0`` is equivalent
    to using ``penalty='l2'``, while setting ``l1_ratio=1`` is equivalent
    to using ``penalty='l1'``. For ``0 < l1_ratio <1``, the penalty is a
    combination of L1 and L2.
"""
Penalty = Literal["l1", "l2", "elasticnet", "none"]
ClassWeight = Optional[dict | Literal["balanced"]]
Solver = Literal["newton-cg", "lbfgs", "liblinear", "sag", "saga"]
MultiClass = Literal["auto", "ovr", "multinomial"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class LogisticRegressorConfig(ModelConfig):
    penalty: Penalty = "l2"
    dual: bool = False
    tol: float = 1e-4
    C: float = 1
    fit_intercept: bool = True
    intercept_scaling: float = 1
    class_weight: ClassWeight = None
    random_state: int | RandomState | None = None
    solver: Solver = "lbfgs"
    max_iter: int = 100
    verbose: int = 0
    warm_start: bool = False
    l1_ratio: float | None = None


class LogisticRegressor(
    SKLearnClassificationMixin,
    SKLearnModelBase[LogisticRegressorConfig],
):
    @property
    def _config_builder(self) -> type[LogisticRegressorConfig]:
        return LogisticRegressorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKLogisticRegression:
        config = self.model_config
        return SKLogisticRegression(
            penalty=config.penalty,
            dual=config.dual,
            tol=config.tol,
            C=config.C,
            fit_intercept=config.fit_intercept,
            intercept_scaling=config.intercept_scaling,
            class_weight=config.class_weight,
            random_state=config.random_state,
            solver=config.solver,
            max_iter=config.max_iter,
            verbose=config.verbose,
            warm_start=config.warm_start,
            l1_ratio=config.l1_ratio,
        )



from typing import Any

try:
    from mhfp.encoder import MHFPEncoder
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
MHFP6 (MinHash fingerprint, up to six bonds) is a molecular fingerprint which
encodes detailed substructures using the extended connectivity principle of
ECFP in a fundamentally different manner, increasing the performance of exact
nearest neighbor searches in benchmarking studies and enabling the application
of locality sensitive hashing (LSH) approximate nearest neighbor search
algorithms.

To describe a molecule, MHFP6 extracts the SMILES of all circular substructures
around each atom up to a diameter of six bonds and applies the MinHash method
to the resulting set. MHFP6 outperforms ECFP4 in benchmarking analog recovery
studies. Furthermore, MHFP6 outperforms ECFP4 in approximate nearest neighbor
searches by two orders of magnitude in terms of speed, while decreasing the
error rate

https://jcheminf.biomedcentral.com/articles/10.1186/s13321-018-0321-8
"""


class MHFP(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        length: int = 2048,
        radius: int = 3,
        rings: bool = True,
        kekulise: bool = True,
        sanitise: bool = True,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Featurises the input molecules as SECFP (SMILES Extended
        Connectifity Fingerprint) MinHash Fingerprint.

        Args:
            samples: The array of samples to featurise.
            length: The length of the folded fingerprint. Defaults to 2048.
            radius: Analogous to the radius for the Morgan fingerprint. The
                default radius 3 encodes SMILES to MHFP6. Defaults to 3.
            rings: Whether rings in the molecule are included in the fingerprint.
                As a radii of 3 fails to encode rings and there is no way to
                determine ring-membership in a substructure SMILES, this
                considerably increases performance. Defaults to `True`.
            kekulise: Whether or not to kekulise the molecule before extracting
                substructure SMILES. Defaults to `True`.
            sanitise: Whether or not to sanitise the molecule (using RDKit)
                before extracting substructure SMILES. Defaults to `True`.

        Returns:
            MHFP fingerpints.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation("mhfp")
            >>> samples = ["CCCC1=NN(C2=C1NC(=NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C"]
            >>> representation.featurise(samples, length=4)
            {'mhfp': [[1, 1, 1, 1]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        fingerprints = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smiles = to_smiles(sample)
                fingerprint = MHFPEncoder.secfp_from_smiles(
                    smiles,
                    length=length,
                    radius=radius,
                    rings=rings,
                    kekulize=kekulise,
                    sanitize=sanitise,
                ).tolist()
                fingerprints.append(fingerprint)

        return {self.tag: fingerprints}



import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.regression.max_error import MaxError


@pytest.fixture()
def fixture_metric():
    return load_metric("max_error")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "max_error" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, MaxError)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_default_compute(fixture_metric):
    """That default scoring."""
    metric = fixture_metric
    predictions = [9, 2, 7, 1]
    references = [3, 2, 7, 1]
    result = metric.compute(predictions=predictions, references=references)
    assert result["max_error"] == 6


def test_perfect_score(fixture_metric):
    """Test score on perfect predictions."""
    metric = fixture_metric
    predictions = [3, 2, 7, 1]
    references = [3, 2, 7, 1]
    result = metric.compute(predictions=predictions, references=references)
    assert result["max_error"] == 0



import warnings

import pandas as pd
import pytest

import datasets
from molflux.modelzoo.load import load_model
from molflux.modelzoo.protocols import Estimator
from molflux.modelzoo.typing import DataFrameLike


def test_load() -> None:
    model = load_model("mock_model")
    assert model is not None


def test_implements_protocol() -> None:
    model = load_model("mock_model")
    assert isinstance(model, Estimator)


train_df = pd.DataFrame({"x1": ["Hello"], "y1": ["Buongiorno"]})


@pytest.mark.parametrize(
    "train_data",
    [
        train_df,
        datasets.Dataset.from_pandas(train_df),
    ],
)
def test_train(train_data: DataFrameLike) -> None:
    model = load_model("mock_model", x_features=["x1"], y_features=["y1"])
    model.train(train_data=train_data)
    assert True


def test_unexpected_validation_data_in_model_train() -> None:
    model = load_model(
        "mock_model",
        x_features=["x1"],
        y_features=["y1"],
    )
    with pytest.warns(UserWarning):
        model.train(train_data=train_df, validation_data=train_df)


def test_expected_validation_data_in_model_train() -> None:
    model = load_model(
        "mock_model_with_validation_data_arg",
        x_features=["x1"],
        y_features=["y1"],
    )
    with warnings.catch_warnings():
        model.train(train_data=train_df, validation_data=train_df)


def test_unexpected_validation_data_in_model_train_multi_data() -> None:
    model = load_model(
        "mock_model_multi_data",
        x_features=["x1"],
        y_features=["y1"],
    )
    dataset: dict[str, DataFrameLike] = {"foo": train_df}
    with pytest.warns(UserWarning):
        model.train(train_data=dataset, validation_data=dataset)


def test_expected_validation_data_in_model_train_multi_data() -> None:
    model = load_model(
        "mock_model_multi_data_with_validation_data_arg",
        x_features=["x1"],
        y_features=["y1"],
    )
    dataset: dict[str, DataFrameLike] = {"foo": train_df}
    with warnings.catch_warnings():
        model.train(train_data=dataset, validation_data=dataset)



"""F1 score, also known as balanced F-score or F-measure"""

import logging
from typing import Any

import evaluate
from sklearn.metrics import f1_score

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Compute the F1 score, also known as balanced F-score or F-measure.
The F1 score can be interpreted as a harmonic mean of the precision and
recall, where an F1 score reaches its best value at 1 and worst score at 0.
The relative contribution of precision and recall to the F1 score are
equal. The formula for the F1 score is::

    F1 = 2 * (precision * recall) / (precision + recall)

In the multi-class and multi-label case, this is the average of
the F1 score of each class with weighting depending on the ``average``
parameter.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated targets as returned by a classifier.
    references: Ground truth (correct) target values.
    average (optional): {'micro', 'macro', 'samples','weighted', 'binary'} or
        None, default='binary'. This parameter is required for
        multiclass/multilabel targets. If ``None``, the scores for each class
        are returned. Otherwise, this determines the type of averaging performed
        on the data:

        ``'binary'``:
            Only report results for the class specified by ``pos_label``.
            This is applicable only if targets (``y_{true,pred}``) are binary.
        ``'micro'``:
            Calculate metrics globally by counting the total true positives,
            false negatives and false positives.
        ``'macro'``:
            Calculate metrics for each label, and find their unweighted
            mean.  This does not take label imbalance into account.
        ``'weighted'``:
            Calculate metrics for each label, and find their average weighted
            by support (the number of true instances for each label). This
            alters 'macro' to account for label imbalance; it can result in an
            F-score that is not between precision and recall.
        ``'samples'``:
            Calculate metrics for each instance, and find their average (only
            meaningful for multilabel classification where this differs from
            :func:`accuracy_score`).
    labels (optional): The set of labels to include when
        ``average != 'binary'``, and their order if ``average is None``.
        Labels present in the data can be excluded, for example to calculate a
        multiclass average ignoring a majority negative class, while labels not
        present in the data will result in 0 components in a macro average.
        For multilabel targets, labels are column indices. By default, all
        labels in ``y_true`` and ``y_pred`` are used in sorted order.
    pos_label (optional): The class to report if ``average='binary'`` and the
        data is binary. If the data are multiclass or multilabel, this will be
        ignored; setting ``labels=[pos_label]`` and ``average != 'binary'``
        will report scores for that label only. Defaults to 1.
    sample_weight (optional): Weighting of each sample.
    zero_division (optional): "warn", 0 or 1. Sets the value to return when
        there is a zero division, i.e. when all predictions and labels are
        negative. If set to "warn", this acts as 0, but warnings are also
        raised. Defaults to "warn".

Returns:
    f1_score: F1 score of the positive class in binary classification or
    weighted average of the F1 scores of each class for the multiclass task.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("f1_score")
    >>> references = [0, 1, 2, 0, 1, 2]
    >>> predictions = [0, 2, 1, 0, 0, 1]
    >>> metric.compute(references=references, predictions=predictions, average="macro")
    {'f1_score': 0.26...}
    >>> metric.compute(references=references, predictions=predictions, average="micro")
    {'f1_score': 0.33...}
    >>> metric.compute(references=references, predictions=predictions, average="weighted")
    {'f1_score': 0.26...}
    >>> metric.compute(references=references, predictions=predictions, average=None)
    {'f1_score': [0.8, 0.0, 0.0]}
    >>> references = [0, 0, 0, 0, 0, 0]
    >>> predictions = [0, 0, 0, 0, 0, 0]
    >>> metric.compute(references=references, predictions=predictions, zero_division=1)
    {'f1_score': 1.0}

Notes:
    When ``true positive + false positive == 0``, precision is undefined.
    When ``true positive + false negative == 0``, recall is undefined.
    In such cases, by default the metric will be set to 0, as will f-score,
    and ``UndefinedMetricWarning`` will be raised. This behavior can be
    modified with ``zero_division``.
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class F1Score(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("int32")),
                    "references": datasets.Sequence(datasets.Value("int32")),
                }
                if self.config_name == "multilabel"
                else {
                    "predictions": datasets.Value("int32"),
                    "references": datasets.Value("int32"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.balanced_accuracy_score.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: list[float] | None = None,
        labels: list[str] | None = None,
        pos_label: int = 1,
        average: str | None = "binary",
        zero_division: int | str = "warn",
        **kwargs: Any,
    ) -> MetricResult:
        score = f1_score(
            y_true=references,
            y_pred=predictions,
            labels=labels,
            pos_label=pos_label,
            average=average,
            sample_weight=sample_weight,
            zero_division=zero_division,
        )
        return {self.tag: score.tolist()}



import pytest

from datasets.builder import DatasetBuilder
from molflux.datasets.catalogue import (
    get_builder_entrypoint,
    list_datasets,
    register_builder,
)


def test_catalogue_is_filled():
    """That the catalogue is not empty."""
    catalogue = list_datasets()
    assert len(catalogue)


def test_register_builder():
    """That a custom class is registered. TODO make this temporary"""
    kind = "test"
    name = "custom_name"

    @register_builder(kind=kind, name=name)
    class MyClass(DatasetBuilder):
        pass

    catalogue = list_datasets()
    assert kind in catalogue
    assert catalogue[kind] == [name]


def test_get_builder_entrypoint():
    with pytest.raises(NotImplementedError):
        get_builder_entrypoint("molflux_endpoint_not_available")



"""
Tests ensuring desired API objects are part of the top-level namespace.
"""

import pytest

import molflux.datasets


@pytest.mark.parametrize(
    "callable_name",
    [
        "fill_catalogue",
        "list_datasets",
        "load_dataset_from_store",
        "save_dataset_to_store",
        "featurise_dataset",
        "load_dataset",
        "load_dataset_builder",
        "load_from_dict",
        "load_from_yaml",
        "split_dataset",
    ],
)
def test_exports_callable(callable_name):
    """That the package exposes the given callable."""
    assert hasattr(molflux.datasets, callable_name)
    assert callable(getattr(molflux.datasets, callable_name))



import difflib
import functools
import inspect
import logging
from collections import defaultdict
from collections.abc import Callable
from importlib.metadata import EntryPoint, entry_points

from molflux.modelzoo.naming import camelcase_to_snakecase
from molflux.modelzoo.protocols import Estimator

logger = logging.getLogger(__name__)

EstimatorT = type[Estimator]

NAMESPACE = "molflux.modelzoo.plugins."

# This is where entrypoints will be registered {<name>: <Entrypoint>}
MODELS_CATALOGUE: dict[str, EntryPoint] = {}


@functools.lru_cache
def fill_catalogue() -> None:
    """Fills the catalogue with entrypoints.

    Only ever call this once for performance reasons.
    """
    entrypoints = [
        entrypoint
        for namespace, entrypoints in entry_points().items()
        if namespace.startswith(NAMESPACE)
        for entrypoint in entrypoints
    ]
    for entrypoint in entrypoints:
        put_model_entrypoint(model_entrypoint=entrypoint)


def put_model_entrypoint(model_entrypoint: EntryPoint) -> None:
    """Puts a representation entrypoint in the catalogue."""
    model_name = camelcase_to_snakecase(model_entrypoint.name)

    if model_name in MODELS_CATALOGUE:
        if MODELS_CATALOGUE[model_name] == model_entrypoint:
            pass
        else:
            raise KeyError(f"Duplicate model: {model_name!r}")

    MODELS_CATALOGUE[model_name] = model_entrypoint


def get_model_entrypoint(model_name: str) -> EntryPoint:
    """Returns a specific model entrypoint from the catalogue.

    The model is keyed only by name. This allows to decouple from
    eventual reshuffles of entrypoints across groups.
    """
    model_entrypoint = MODELS_CATALOGUE.get(model_name)
    if model_entrypoint is None:
        msg = f"Model {model_name!r} is not available."
        similar = difflib.get_close_matches(model_name, MODELS_CATALOGUE.keys())
        if similar:
            msg += f" You might be looking for one of these: {similar}"
        raise NotImplementedError(msg)
    return model_entrypoint


def get_model_cls(model_name: str) -> EstimatorT:
    entrypoint = get_model_entrypoint(model_name=model_name)

    # Might raise (e.g. missing dependency extras)
    model_cls = entrypoint.load()

    if not isinstance(model_cls, Estimator):
        raise NotImplementedError(
            f"Plugin {entrypoint.value} does not implement Model protocol.",
        )

    return model_cls  # type: ignore[return-value]


def list_models() -> dict[str, list[str]]:
    """List all available models in the catalogue.

    The catalogue is returned as a view of model names keyed by kind.
    """

    view: dict[str, list[str]] = defaultdict(list)

    for model_name, model_entrypoint in MODELS_CATALOGUE.items():
        model_kind = model_entrypoint.group.split(".")[-1]
        view[model_kind].append(model_name)

    return dict(sorted(view.items()))


def register_model(kind: str, name: str) -> Callable:
    """Registers a custom model in the entrypoints catalogue.

    Examples:
        .. code-block:: python

           @register_model(kind="custom", name="my_model")
           class CustomModel:
               ...
    """

    def wrapper(model_cls: EstimatorT) -> EstimatorT:
        module = inspect.getmodule(model_cls).__name__  # type: ignore
        value = f"{module}:{model_cls.__name__}"
        group = NAMESPACE + kind
        entrypoint = EntryPoint(name=name, value=value, group=group)
        put_model_entrypoint(model_entrypoint=entrypoint)
        return model_cls

    return wrapper



from typing import Any

import pytest

import datasets
import molflux.modelzoo
from molflux.core import featurise_dataset, replay_dataset_featurisation, save_model


@pytest.fixture()
def fixture_sample_dataset() -> datasets.Dataset:
    """A sample training dataset. Smiles from TDC lipo"""
    return datasets.Dataset.from_dict(
        {
            "canonical_smiles": [
                "CCCCCCCC(=O)OCN1C(=O)C(NC1=O)(c2ccccc2)c3ccccc3",
                "CCN(=O)=O",
                "CCN(CC(=C)C)c1c(cc(cc1N(=O)=O)C(F)(F)F)N(=O)=O",
                "c1cc(c(c(c1Cl)Cl)Cl)Cl",
                "CCCC(C)(COC(=O)N)COC(=O)N",
            ],
            "y": [1, 2, 3, 4, 5],
        },
    )


@pytest.fixture()
def fixture_sample_featurisation_metadata_v1() -> dict[str, Any]:
    """A sample v1 featurisation metadata for a dataset with 'canonical_smiles'"""
    return {
        "version": 1,
        "config": [
            {
                "column": "canonical_smiles",
                "representations": [
                    {
                        "name": "character_count",
                        "as": "character_count",
                    },
                    {
                        "name": "character_count",
                        "as": "my_alias",
                        "config": {
                            "tag": "c_count",
                        },
                        "presets": {},
                    },
                ],
            },
        ],
    }


@pytest.fixture()
def fixture_sample_featurisation_metadata_v2() -> dict[str, Any]:
    """A sample v2 featurisation metadata for a dataset with 'canonical_smiles'"""
    return {
        "version": 2,
        "config": [
            {
                "columns": ["canonical_smiles"],
                "representations": [
                    {
                        "name": "character_count",
                        "as": "character_count",
                    },
                    {
                        "name": "character_count",
                        "as": "my_alias",
                        "config": {
                            "tag": "c_count",
                        },
                        "presets": {},
                    },
                ],
            },
            {
                "columns": ["character_count", "my_alias"],
                "representations": [
                    {
                        "as": "character_count_plus_my_alias",
                        "name": "sum",
                    },
                ],
            },
        ],
    }


def test_featurise_dataset_with_unsupported_featurisation_metadata_version_raises(
    fixture_sample_dataset,
):
    """That an error is raised if attempting to featurise a dataset with an
    unsupported version of featurisation metadata schema."""

    dataset = fixture_sample_dataset
    featurisation_metadata = {"version": 0, "config": {}}

    with pytest.raises(NotImplementedError, match="version"):
        featurise_dataset(dataset, featurisation_metadata=featurisation_metadata)


@pytest.mark.parametrize(
    ["fixture_sample_featurisation_metadata", "expected_cols"],
    (
        [
            "fixture_sample_featurisation_metadata_v1",
            ["my_alias"],
        ],
        [
            "fixture_sample_featurisation_metadata_v2",
            ["character_count", "my_alias", "character_count_plus_my_alias"],
        ],
    ),
)
def test_featurise_dataset(
    fixture_sample_dataset,
    fixture_sample_featurisation_metadata,
    expected_cols,
    request,
):
    """That can featurise a dataset using well-formed featurisation metadata."""

    dataset = fixture_sample_dataset
    featurisation_metadata = request.getfixturevalue(
        fixture_sample_featurisation_metadata,
    )

    featurised_dataset = featurise_dataset(
        dataset,
        featurisation_metadata=featurisation_metadata,
    )
    assert len(featurised_dataset.column_names) > len(dataset.column_names)
    for col in expected_cols:
        assert col in featurised_dataset.column_names


@pytest.mark.parametrize(
    ["fixture_sample_featurisation_metadata", "expected_cols"],
    (
        [
            "fixture_sample_featurisation_metadata_v1",
            ["my_alias"],
        ],
        [
            "fixture_sample_featurisation_metadata_v2",
            ["character_count", "my_alias", "character_count_plus_my_alias"],
        ],
    ),
)
def test_featurise_dataset_with_map_kwargs(
    fixture_sample_dataset,
    fixture_sample_featurisation_metadata,
    expected_cols,
    request,
):
    """That can featurise a dataset using well-formed featurisation metadata and use map_kwargs."""

    dataset = fixture_sample_dataset
    featurisation_metadata = request.getfixturevalue(
        fixture_sample_featurisation_metadata,
    )

    featurised_dataset = featurise_dataset(
        dataset,
        featurisation_metadata=featurisation_metadata,
        batch_size=2,
    )
    assert len(featurised_dataset.column_names) > len(dataset.column_names)
    for col in expected_cols:
        assert col in featurised_dataset.column_names


@pytest.mark.parametrize(
    ["fixture_sample_featurisation_metadata", "x_features"],
    (
        [
            "fixture_sample_featurisation_metadata_v1",
            ["my_alias"],
        ],
        [
            "fixture_sample_featurisation_metadata_v2",
            ["character_count", "my_alias", "character_count_plus_my_alias"],
        ],
    ),
)
def test_replay_featurisation(
    tmp_path,
    fixture_sample_dataset,
    fixture_sample_featurisation_metadata,
    x_features,
    request,
):
    """That can replay featurisation for a given saved model."""

    dataset = fixture_sample_dataset
    featurisation_metadata = request.getfixturevalue(
        fixture_sample_featurisation_metadata,
    )
    featurised_dataset = featurise_dataset(
        dataset,
        featurisation_metadata=featurisation_metadata,
    )

    model = molflux.modelzoo.load_model(
        "linear_regressor",
        x_features=x_features,
        y_features=["y"],
    )
    model.train(featurised_dataset)
    model_path = save_model(
        model,
        path=tmp_path,
        featurisation_metadata=featurisation_metadata,
    )

    replayed_dataset = replay_dataset_featurisation(dataset, model_path=model_path)
    assert replayed_dataset.column_names == featurised_dataset.column_names


@pytest.mark.parametrize(
    ["fixture_sample_featurisation_metadata", "x_features"],
    (
        [
            "fixture_sample_featurisation_metadata_v1",
            ["my_alias"],
        ],
        [
            "fixture_sample_featurisation_metadata_v2",
            ["character_count", "my_alias", "character_count_plus_my_alias"],
        ],
    ),
)
def test_replay_featurisation_with_map_kwargs(
    tmp_path,
    fixture_sample_dataset,
    fixture_sample_featurisation_metadata,
    x_features,
    request,
):
    """That can replay featurisation for a given saved model with map_kwargs."""

    dataset = fixture_sample_dataset
    featurisation_metadata = request.getfixturevalue(
        fixture_sample_featurisation_metadata,
    )
    featurised_dataset = featurise_dataset(
        dataset,
        featurisation_metadata=featurisation_metadata,
    )

    model = molflux.modelzoo.load_model(
        "linear_regressor",
        x_features=x_features,
        y_features=["y"],
    )
    model.train(featurised_dataset)
    model_path = save_model(
        model,
        path=tmp_path,
        featurisation_metadata=featurisation_metadata,
    )

    replayed_dataset = replay_dataset_featurisation(
        dataset,
        model_path=model_path,
        batch_size=2,
    )
    assert replayed_dataset.column_names == featurised_dataset.column_names



import pytest

from molflux.metrics.errors import DuplicateKeyError
from molflux.metrics.load import (
    load_from_dict,
    load_from_dicts,
    load_from_yaml,
    load_metric,
    load_metrics,
)
from molflux.metrics.metric import Metric, Metrics

representative_metric_name = "explained_variance"


def test_returns_metric():
    """That loading a metric returns an object of type Metric."""
    name = representative_metric_name
    metric = load_metric(name=name)
    assert isinstance(metric, Metric)


def test_raises_not_implemented_for_unknown_metric():
    """That a NotImplementedError is raised if attempting to load an unknown
    metric."""
    name = "unknown"
    with pytest.raises(NotImplementedError):
        load_metric(name=name)


def test_raised_error_for_unknown_metric_provides_close_matches():
    """That the error raised when attempting to load an unknown metric
    shows possible close matches to the user (if any)."""
    name = "accurcy"
    # This should suggest e.g. ["accuracy", "top_k_accuracy"]
    with pytest.raises(
        NotImplementedError,
        match="You might be looking for one of these",
    ):
        load_metric(name=name)


def test_forwards_init_kwargs_to_builder():
    """That keyword arguments get forwarded to the metric initialiser."""
    name = representative_metric_name
    metric = load_metric(name=name, tag="pytest-tag")
    assert metric.tag == "pytest-tag"


def test_metrics_with_same_tag_in_collection_raises():
    """That adding several metrics with the same tag in a Metrics collection raises."""
    name = representative_metric_name
    with pytest.raises(DuplicateKeyError):
        load_metrics(name, name)


def test_load_metrics_returns_metrics():
    """That the load_metrics function returns a Metrics collection."""
    name = representative_metric_name
    metrics = load_metrics(name, name, tags=["one", "two"])
    assert isinstance(metrics, Metrics)


def test_load_metrics_returns_correct_number_of_metrics():
    """That the load_metrics function returns a Metrics collection of the
    expected size."""
    name = representative_metric_name
    metrics = load_metrics(name, name, tags=["one", "two"])
    assert len(metrics) == 2


def test_from_dict_returns_metric():
    """That loading from a dict returns a Metric."""
    name = representative_metric_name
    config = {
        "name": name,
        "config": {},
        "presets": {},
    }
    metric = load_from_dict(config)
    assert isinstance(metric, Metric)


def test_from_minimal_dict():
    """That can provide a config with only required fields."""
    name = representative_metric_name
    config = {
        "name": name,
    }
    assert load_from_dict(config)


def test_from_dict_forwards_config_to_builder():
    """That config keyword arguments get forwarded to the initialiser."""
    name = representative_metric_name
    config = {
        "name": name,
        "config": {
            "tag": "pytest-tag",
        },
    }
    metric = load_from_dict(config)
    assert metric.tag == "pytest-tag"


def test_from_dict_forwards_presets_to_state():
    """That presets keyword arguments get stored in the metric state."""
    name = representative_metric_name
    config = {
        "name": name,
        "presets": {
            "multioutput": "variance_weighted",
        },
    }
    metric = load_from_dict(config)
    assert metric.state
    assert metric.state.get("multioutput") == "variance_weighted"


def test_dict_missing_required_fields_raises():
    """That cannot load a metric with a config missing required fields."""
    config = {"unknown_key": "value"}
    with pytest.raises(SyntaxError):
        load_from_dict(config)


def test_from_dicts_returns_metrics():
    """That loading from a collection of dicts returns a Metrics object."""
    name = representative_metric_name
    config_one = {
        "name": name,
        "config": {
            "tag": "one",
        },
        "presets": {},
    }
    config_two = {
        "name": name,
        "config": {
            "tag": "two",
        },
        "presets": {},
    }
    configs = [config_one, config_two]
    metrics = load_from_dicts(configs)
    assert isinstance(metrics, Metrics)


def test_from_yaml_returns_metrics(fixture_path_to_assets):
    path = fixture_path_to_assets / "config.yml"
    metrics = load_from_yaml(path=path)
    assert len(metrics) == 2
    assert "explained_variance" in metrics
    assert "custom_metric" in metrics
    assert isinstance(metrics, Metrics)
    assert metrics["custom_metric"].state.get("root") is True
